import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestTakeLast(unittest.TestCase):
    def test_take_last_zero_completed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(270, 4),
            on_next(310, 5),
            on_next(360, 6),
            on_next(380, 7),
            on_next(410, 8),
            on_next(590, 9),
            on_completed(650),
        )

        def create():
            return xs.pipe(ops.take_last(0))

        results = scheduler.start(create)

        assert results.messages == [on_completed(650)]
        assert xs.subscriptions == [subscribe(200, 650)]

    def test_take_last_zero_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(270, 4),
            on_next(310, 5),
            on_next(360, 6),
            on_next(380, 7),
            on_next(410, 8),
            on_next(590, 9),
            on_error(650, ex),
        )

        def create():
            return xs.pipe(ops.take_last(0))

        results = scheduler.start(create)

        assert results.messages == [on_error(650, ex)]
        assert xs.subscriptions == [subscribe(200, 650)]

    def test_take_last_zero_disposed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(270, 4),
            on_next(310, 5),
            on_next(360, 6),
            on_next(380, 7),
            on_next(410, 8),
            on_next(590, 9),
        )

        def create():
            return xs.pipe(ops.take_last(0))

        results = scheduler.start(create)

        assert results.messages == []
        assert xs.subscriptions == [subscribe(200, 1000)]

    def test_take_last_one_completed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(270, 4),
            on_next(310, 5),
            on_next(360, 6),
            on_next(380, 7),
            on_next(410, 8),
            on_next(590, 9),
            on_completed(650),
        )

        def create():
            return xs.pipe(ops.take_last(1))

        results = scheduler.start(create)
        assert results.messages == [on_next(650, 9), on_completed(650)]
        assert xs.subscriptions == [subscribe(200, 650)]

    def test_take_last_one_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(270, 4),
            on_next(310, 5),
            on_next(360, 6),
            on_next(380, 7),
            on_next(410, 8),
            on_next(590, 9),
            on_error(650, ex),
        )

        def create():
            return xs.pipe(ops.take_last(1))

        results = scheduler.start(create)

        assert results.messages == [on_error(650, ex)]
        assert xs.subscriptions == [subscribe(200, 650)]

    def test_take_last_One_disposed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(270, 4),
            on_next(310, 5),
            on_next(360, 6),
            on_next(380, 7),
            on_next(410, 8),
            on_next(590, 9),
        )

        def create():
            return xs.pipe(ops.take_last(1))

        results = scheduler.start(create)

        assert results.messages == []
        assert xs.subscriptions == [subscribe(200, 1000)]

    def test_take_last_three_completed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(270, 4),
            on_next(310, 5),
            on_next(360, 6),
            on_next(380, 7),
            on_next(410, 8),
            on_next(590, 9),
            on_completed(650),
        )

        def create():
            return xs.pipe(ops.take_last(3))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(650, 7),
            on_next(650, 8),
            on_next(650, 9),
            on_completed(650),
        ]
        assert xs.subscriptions == [subscribe(200, 650)]

    def test_Take_last_three_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(270, 4),
            on_next(310, 5),
            on_next(360, 6),
            on_next(380, 7),
            on_next(410, 8),
            on_next(590, 9),
            on_error(650, ex),
        )

        def create():
            return xs.pipe(ops.take_last(3))

        results = scheduler.start(create)

        assert results.messages == [on_error(650, ex)]
        assert xs.subscriptions == [subscribe(200, 650)]

    def test_Take_last_three_disposed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(270, 4),
            on_next(310, 5),
            on_next(360, 6),
            on_next(380, 7),
            on_next(410, 8),
            on_next(590, 9),
        )

        def create():
            return xs.pipe(ops.take_last(3))

        results = scheduler.start(create)

        assert results.messages == []
        assert xs.subscriptions == [subscribe(200, 1000)]



import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(ex):
    raise RxException(ex)


class TestGenerate(unittest.TestCase):
    def test_generate_finite(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.generate(
                0,
                lambda x: x <= 3,
                lambda x: x + 1,
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(200, 0),
            on_next(200, 1),
            on_next(200, 2),
            on_next(200, 3),
            on_completed(200),
        ]

    def test_generate_throw_condition(self):
        scheduler = TestScheduler()
        ex = "ex"

        def create():
            return reactivex.generate(
                0,
                lambda x: _raise("ex"),
                lambda x: x + 1,
            )

        results = scheduler.start(create)

        assert results.messages == [on_error(200, ex)]

    def test_generate_throw_iterate(self):
        scheduler = TestScheduler()
        ex = "ex"

        def create():
            return reactivex.generate(
                0,
                lambda x: True,
                lambda x: _raise(ex),
            )

        results = scheduler.start(create)

        assert results.messages == [on_next(200, 0), on_error(200, ex)]

    def test_generate_dispose(self):
        scheduler = TestScheduler()
        ex = "ex"

        def create():
            return reactivex.generate(
                0,
                lambda x: True,
                lambda x: x + 1,
            )

        results = scheduler.start(create, disposed=200)
        assert results.messages == []

    def test_generate_repeat(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.generate(
                0,
                lambda x: x <= 3,
                lambda x: x + 1,
            ).pipe(ops.repeat(2))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(200, 0),
            on_next(200, 1),
            on_next(200, 2),
            on_next(200, 3),
            on_next(200, 0),
            on_next(200, 1),
            on_next(200, 2),
            on_next(200, 3),
            on_completed(200),
        ]



import unittest
from datetime import datetime, timedelta

import pytest

from reactivex.scheduler.eventloop import GEventScheduler

gevent = pytest.importorskip("gevent")


class TestGEventScheduler(unittest.TestCase):
    def test_gevent_schedule_now(self):
        scheduler = GEventScheduler(gevent)
        hub = gevent.get_hub()
        diff = scheduler.now - datetime.utcfromtimestamp(hub.loop.now())
        assert abs(diff) < timedelta(milliseconds=1)

    def test_gevent_schedule_now_units(self):
        scheduler = GEventScheduler(gevent)
        diff = scheduler.now
        gevent.sleep(0.1)
        diff = scheduler.now - diff
        assert timedelta(milliseconds=80) < diff < timedelta(milliseconds=180)

    def test_gevent_schedule_action(self):
        scheduler = GEventScheduler(gevent)
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        scheduler.schedule(action)

        gevent.sleep(0.1)
        assert ran is True

    def test_gevent_schedule_action_due(self):
        scheduler = GEventScheduler(gevent)
        starttime = datetime.now()
        endtime = None

        def action(scheduler, state):
            nonlocal endtime
            endtime = datetime.now()

        scheduler.schedule_relative(0.2, action)

        gevent.sleep(0.3)
        assert endtime is not None
        diff = endtime - starttime
        assert diff > timedelta(seconds=0.18)

    def test_gevent_schedule_action_cancel(self):
        scheduler = GEventScheduler(gevent)
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        d = scheduler.schedule_relative(0.01, action)
        d.dispose()

        gevent.sleep(0.1)
        assert ran is False



import logging
import unittest
from datetime import datetime

from reactivex.operators import delay
from reactivex.testing import ReactiveTest, TestScheduler

FORMAT = "%(asctime)-15s %(threadName)s %(message)s"
logging.basicConfig(filename="rx.log", format=FORMAT, level=logging.DEBUG)
log = logging.getLogger("Rx")

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(ex):
    raise RxException(ex)


class TestDelay(unittest.TestCase):
    def test_delay_timespan_simple1(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(250, 2),
            on_next(350, 3),
            on_next(450, 4),
            on_completed(550),
        )

        def create():
            return xs.pipe(delay(100.0))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(350, 2),
            on_next(450, 3),
            on_next(550, 4),
            on_completed(650),
        ]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_datetime_offset_simple1_impl(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(250, 2),
            on_next(350, 3),
            on_next(450, 4),
            on_completed(550),
        )

        def create():
            dt = datetime.utcfromtimestamp(300.0)
            return xs.pipe(delay(dt))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(350, 2),
            on_next(450, 3),
            on_next(550, 4),
            on_completed(650),
        ]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_timespan_simple2_impl(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(250, 2),
            on_next(350, 3),
            on_next(450, 4),
            on_completed(550),
        )

        def create():
            return xs.pipe(delay(50))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(300, 2),
            on_next(400, 3),
            on_next(500, 4),
            on_completed(600),
        ]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_datetime_offset_simple2_impl(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(250, 2),
            on_next(350, 3),
            on_next(450, 4),
            on_completed(550),
        )

        def create():
            return xs.pipe(delay(datetime.utcfromtimestamp(250)))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(300, 2),
            on_next(400, 3),
            on_next(500, 4),
            on_completed(600),
        ]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_timespan_simple3_impl(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(250, 2),
            on_next(350, 3),
            on_next(450, 4),
            on_completed(550),
        )

        def create():
            return xs.pipe(delay(150))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(400, 2),
            on_next(500, 3),
            on_next(600, 4),
            on_completed(700),
        ]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_datetime_offset_simple3_impl(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(250, 2),
            on_next(350, 3),
            on_next(450, 4),
            on_completed(550),
        )

        def create():
            return xs.pipe(delay(datetime.utcfromtimestamp(350)))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(400, 2),
            on_next(500, 3),
            on_next(600, 4),
            on_completed(700),
        ]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_timespan_error1_impl(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(250, 2),
            on_next(350, 3),
            on_next(450, 4),
            on_error(550, ex),
        )

        def create():
            return xs.pipe(delay(50))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(300, 2),
            on_next(400, 3),
            on_next(500, 4),
            on_error(550, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_datetime_offset_error1_impl(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(250, 2),
            on_next(350, 3),
            on_next(450, 4),
            on_error(550, ex),
        )

        def create():
            return xs.pipe(delay(datetime.utcfromtimestamp(250)))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(300, 2),
            on_next(400, 3),
            on_next(500, 4),
            on_error(550, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_timespan_error2_impl(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(250, 2),
            on_next(350, 3),
            on_next(450, 4),
            on_error(550, ex),
        )

        def create():
            return xs.pipe(delay(150))

        results = scheduler.start(create)

        assert results.messages == [on_next(400, 2), on_next(500, 3), on_error(550, ex)]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_datetime_offset_error2_impl(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(250, 2),
            on_next(350, 3),
            on_next(450, 4),
            on_error(550, ex),
        )

        def create():
            return xs.pipe(delay(datetime.utcfromtimestamp(350)))

        results = scheduler.start(create)

        assert results.messages == [on_next(400, 2), on_next(500, 3), on_error(550, ex)]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(550))

        def create():
            return xs.pipe(delay(10))

        results = scheduler.start(create)

        assert results.messages == [on_completed(560)]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_error(550, ex))

        def create():
            return xs.pipe(delay(10.0))

        results = scheduler.start(create)

        assert results.messages == [on_error(550, ex)]
        assert xs.subscriptions == [subscribe(200, 550)]

    def test_delay_never(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1))

        def create():
            return xs.pipe(delay(10))

        results = scheduler.start(create)

        assert results.messages == []
        assert xs.subscriptions == [subscribe(200, 1000)]



import unittest
from datetime import datetime

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class Timestamp(object):
    def __init__(self, value, timestamp):
        if isinstance(timestamp, datetime):
            timestamp = timestamp - datetime.utcfromtimestamp(0)
            timestamp = int(
                timestamp.seconds
            )  # FIXME: Must fix when tests run at fraction of seconds.

        self.value = value
        self.timestamp = timestamp

    def __str__(self):
        return "%s@%s" % (self.value, self.timestamp)

    def equals(self, other):
        return other.timestamp == self.timestamp and other.value == self.value


class TestTimeInterval(unittest.TestCase):
    def test_timestamp_regular(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(230, 3),
            on_next(260, 4),
            on_next(300, 5),
            on_next(350, 6),
            on_completed(400),
        )

        def create():
            def mapper(x):
                return Timestamp(x.value, x.timestamp)

            return xs.pipe(
                ops.timestamp(),
                ops.map(mapper),
            )

        results = scheduler.start(create)
        assert results.messages == [
            on_next(210, Timestamp(2, 210)),
            on_next(230, Timestamp(3, 230)),
            on_next(260, Timestamp(4, 260)),
            on_next(300, Timestamp(5, 300)),
            on_next(350, Timestamp(6, 350)),
            on_completed(400),
        ]

    def test_timestamp_empty(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.empty().pipe(ops.timestamp())

        results = scheduler.start(create)
        assert results.messages == [on_completed(200)]

    def test_timestamp_error(self):
        ex = "ex"
        scheduler = TestScheduler()

        def create():
            return reactivex.throw(ex).pipe(ops.timestamp())

        results = scheduler.start(create)
        assert results.messages == [on_error(200, ex)]

    def test_timestamp_never(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.never().pipe(ops.timestamp())

        results = scheduler.start(create)
        assert results.messages == []



from reactivex import Observer
from reactivex.notification import OnCompleted, OnError, OnNext, from_notifier


class MyObserver(Observer):
    def __init__(self):
        super().__init__()
        self.has_on_next = None
        self.has_on_completed = None
        self.has_on_error = None

    def _on_next_core(self, value):
        self.has_on_next = value

    def _on_error_core(self, error: Exception):
        self.has_on_error = str(error)

    def _on_completed_core(self):
        self.has_on_completed = True


def test_to_observer_notification_on_next():
    i = 0

    def next(n):
        assert i == 0
        assert n.kind == "N"
        assert n.value == 42
        assert not hasattr(n, "exception")
        assert n.has_value

    from_notifier(next).on_next(42)


def test_to_observer_notification_on_error():
    ex = "ex"
    i = 0

    def next(n):
        assert i == 0
        assert n.kind == "E"
        assert str(n.exception) == ex
        assert not n.has_value

    from_notifier(next).on_error(ex)


def test_to_observer_notification_completed():
    i = 0

    def next(n):
        assert i == 0
        assert n.kind == "C"
        assert not n.has_value

    from_notifier(next).on_completed()


def test_to_notifier_forwards():
    obsn = MyObserver()
    obsn.to_notifier()(OnNext(42))
    assert obsn.has_on_next == 42

    ex = "ex"
    obse = MyObserver()
    obse.to_notifier()(OnError(ex))
    assert ex == obse.has_on_error

    obsc = MyObserver()
    obsc.to_notifier()(OnCompleted())
    assert obsc.has_on_completed


def test_create_on_next():
    next = [False]

    def on_next(x):
        assert 42 == x
        next[0] = True

    res = Observer(on_next)

    res.on_next(42)
    assert next[0]
    return res.on_completed()


def test_create_on_next_has_error():
    ex = "ex"
    next = [False]
    _e = None

    def on_next(x):
        assert 42 == x
        next[0] = True

    res = Observer(on_next)

    res.on_next(42)
    assert next[0]

    try:
        res.on_error(ex)
        assert False
    except Exception as e:
        e_ = e.args[0]

    assert ex == e_


def test_create_on_next_on_completed():
    next = [False]
    completed = [False]

    def on_next(x):
        assert 42 == x
        next[0] = True
        return next[0]

    def on_completed():
        completed[0] = True
        return completed[0]

    res = Observer(on_next, None, on_completed)

    res.on_next(42)

    assert next[0]
    assert not completed[0]

    res.on_completed()

    assert completed[0]


def test_create_on_next_close_has_error():
    e_ = None
    ex = "ex"
    next = [False]
    completed = [False]

    def on_next(x):
        assert 42 == x
        next[0] = True

    def on_completed():
        completed[0] = True

    res = Observer(on_next, None, on_completed)

    res.on_next(42)
    assert next[0]
    assert not completed[0]
    try:
        res.on_error(ex)
        assert False
    except Exception as e:
        e_ = e.args[0]

    assert ex == e_
    assert not completed[0]


def test_create_on_next_on_error():
    ex = "ex"
    next = [True]
    error = [False]

    def on_next(x):
        assert 42 == x
        next[0] = True

    def on_error(e):
        assert ex == e
        error[0] = True

    res = Observer(on_next, on_error)

    res.on_next(42)

    assert next[0]
    assert not error[0]

    res.on_error(ex)
    assert error[0]


def test_create_on_next_throw_hit_completed():
    ex = "ex"
    next = [True]
    error = [False]

    def on_next(x):
        assert 42 == x
        next[0] = True

    def on_error(e):
        assert ex == e
        error[0] = True

    res = Observer(on_next, on_error)

    res.on_next(42)
    assert next[0]
    assert not error[0]

    res.on_completed()

    assert not error[0]


def test_create_on_next_throw_close1():
    ex = "ex"
    next = [True]
    error = [False]
    completed = [False]

    def on_next(x):
        assert 42 == x
        next[0] = True

    def on_error(e):
        assert ex == e
        error[0] = True

    def on_completed():
        completed[0] = True

    res = Observer(on_next, on_error, on_completed)

    res.on_next(42)

    assert next[0]
    assert not error[0]
    assert not completed[0]

    res.on_completed()

    assert completed[0]
    assert not error[0]


def test_create_on_next_throw_close2():
    ex = "ex"
    next = [True]
    error = [False]
    completed = [False]

    def on_next(x):
        assert 42 == x
        next[0] = True

    def on_error(e):
        assert ex == e
        error[0] = True

    def on_completed():
        completed[0] = True

    res = Observer(on_next, on_error, on_completed)

    res.on_next(42)

    assert next[0]
    assert not error[0]
    assert not completed[0]

    res.on_error(ex)

    assert not completed[0]
    assert error[0]


def test_as_observer_hides():
    obs = MyObserver()
    res = obs.as_observer()

    assert res != obs
    assert not isinstance(res, obs.__class__)


def test_as_observer_forwards():
    obsn = MyObserver()
    obsn.as_observer().on_next(42)
    assert obsn.has_on_next == 42

    ex = "ex"
    obse = MyObserver()
    obse.as_observer().on_error(ex)
    assert obse.has_on_error == ex

    obsc = MyObserver()
    obsc.as_observer().on_completed()
    assert obsc.has_on_completed


if __name__ == "__main__":
    test_to_notifier_forwards()



import unittest

from reactivex.internal.constants import DELTA_ZERO, UTC_ZERO
from reactivex.scheduler.scheduler import Scheduler


class TestScheduler(unittest.TestCase):
    def test_base_to_seconds(self):
        val = Scheduler.to_seconds(0.0)
        assert val == 0.0
        val = Scheduler.to_seconds(DELTA_ZERO)
        assert val == 0.0
        val = Scheduler.to_seconds(UTC_ZERO)
        assert val == 0.0

    def test_base_to_datetime(self):
        val = Scheduler.to_datetime(0.0)
        assert val == UTC_ZERO
        val = Scheduler.to_datetime(DELTA_ZERO)
        assert val == UTC_ZERO
        val = Scheduler.to_datetime(UTC_ZERO)
        assert val == UTC_ZERO

    def test_base_to_timedelta(self):
        val = Scheduler.to_timedelta(0.0)
        assert val == DELTA_ZERO
        val = Scheduler.to_timedelta(DELTA_ZERO)
        assert val == DELTA_ZERO
        val = Scheduler.to_timedelta(UTC_ZERO)
        assert val == DELTA_ZERO



import concurrent.futures
import time

import reactivex
from reactivex import operators as ops

seconds = [5, 1, 2, 4, 3]


def sleep(tm: float) -> float:
    time.sleep(tm)
    return tm


def output(result: str) -> None:
    print("%d seconds" % result)


with concurrent.futures.ProcessPoolExecutor(5) as executor:
    reactivex.from_(seconds).pipe(
        ops.flat_map(lambda s: executor.submit(sleep, s))
    ).subscribe(output)

# 1 seconds
# 2 seconds
# 3 seconds
# 4 seconds
# 5 seconds



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/stable/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
from distutils.command.config import config

import guzzle_sphinx_theme
import tomli
from dunamai import Version

root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.insert(0, root)


# -- Project information -----------------------------------------------------

# General project metadata is stored in pyproject.toml
with open(os.path.join(root, "pyproject.toml"), "rb") as f:
    config = tomli.load(f)

project_meta = config["tool"]["poetry"]

print(project_meta)
project = project_meta["name"]
author = project_meta["authors"][0]
description = project_meta["description"]
url = project_meta["homepage"]
title = project + " Documentation"

_version = Version.from_git()
# The full version, including alpha/beta/rc tags
release = _version.serialize(metadata=False)
# The short X.Y.Z version
version = _version.base


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = "2.0"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "guzzle_sphinx_theme",
    "sphinxcontrib_dooble",
]

# Include a separate entry for special methods, like __init__, where provided.
autodoc_default_options = {
    "member-order": "bysource",
    "special-members": True,
    "exclude-members": "__dict__,__weakref__",
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_translator_class = "guzzle_sphinx_theme.HTMLTranslator"
html_theme_path = guzzle_sphinx_theme.html_theme_path()
html_theme = "guzzle_sphinx_theme"
html_title = title
html_short_title = project + " " + version

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}
html_theme_options = {"projectlink": url}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#

html_sidebars = {"**": ["logo-text.html", "globaltoc.html", "searchbox.html"]}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = project + "doc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [(master_doc, project + ".tex", title, author, "manual")]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, project.lower(), title, [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, project, title, author, project, description, "Miscellaneous")
]


# -- Extension configuration -------------------------------------------------



"""
The automcomplete example rewritten for bottle / gevent.
- Requires besides bottle and gevent also the geventwebsocket pip package
- Instead of a future we create the inner stream for flat_map_latest manually
"""
import json

import gevent
import requests
from bottle import Bottle, abort, request
from geventwebsocket import WebSocketError
from geventwebsocket.handler import WebSocketHandler

from reactivex.scheduler.eventloop import GEventScheduler
from reactivex.subject import Subject


class WikiFinder:
    tmpl = "http://en.wikipedia.org/w/api.php"
    tmpl += "?action=opensearch&search=%s&format=json"

    def __init__(self, term):
        self.res = r = gevent.event.AsyncResult()
        gevent.spawn(lambda: requests.get(self.tmpl % term).text).link(r)

    def subscribe(self, on_next, on_err, on_compl):
        try:
            self.res.get()
            on_next(self.res.value)
        except Exception as ex:
            on_err(ex.args)
        on_compl()


app, PORT = Bottle(), 8081
scheduler = GEventScheduler(gevent)


@app.route("/ws")
def handle_websocket():

    wsock = request.environ.get("wsgi.websocket")
    if not wsock:
        abort(400, "Expected WebSocket request.")
    stream = Subject()
    query = (
        stream.map(lambda x: x["term"])
        .filter(lambda text: len(text) > 2)  # Only if text is longer than 2 characters
        .debounce(0.750, scheduler=scheduler)  # Pause for 750ms
        .distinct_until_changed()
    )  # Only if the value has changed

    searcher = query.flat_map_latest(lambda term: WikiFinder(term))

    def send_response(x):
        wsock.on_next(x)

    def on_error(ex):
        print(ex)

    searcher.subscribe(send_response, on_error)

    while True:
        try:
            message = wsock.receive()
            # like {'term': '<current textbox val>'}
            obj = json.loads(message)
            stream.on_next(obj)
        except WebSocketError:
            break


@app.route("/static/autocomplete.js")
def get_js():
    # blatantly ignoring bottle's template engine:
    return open("autocomplete.js").read().replace("8080", str(PORT))


@app.route("/")
def get_index():
    return open("index.html").read()


if __name__ == "__main__":
    h = ("0.0.0.0", PORT)
    server = gevent.pywsgi.WSGIServer(h, app, handler_class=WebSocketHandler)
    server.serve_forever()



import reactivex
import reactivex.operators as ops

"""
Use a dictionnary to convert elements declared in the marbles diagram to
the specified values.
"""

lookup0 = {"a": 1, "b": 3, "c": 5}
lookup1 = {"x": 2, "y": 4, "z": 6}
source0 = reactivex.cold("a---b----c----|", timespan=0.01, lookup=lookup0)
source1 = reactivex.cold("---x---y---z--|", timespan=0.01, lookup=lookup1)

observable = reactivex.merge(source0, source1).pipe(ops.to_iterable())
elements = observable.run()
print("received {}".format(list(elements)))



import unittest

import reactivex
from reactivex.testing import ReactiveTest

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(ex):
    raise RxException(ex)


class TestFromCallback(unittest.TestCase):
    def test_from_callback(self):
        res = reactivex.from_callback(lambda cb: cb(True))()

        def on_next(r):
            self.assertEqual(r, True)

        def on_error(err):
            assert False

        def on_completed():
            assert True

        res.subscribe(on_next, on_error, on_completed)

    def test_from_callback_single(self):
        res = reactivex.from_callback(lambda file, cb: cb(file))("file.txt")

        def on_next(r):
            self.assertEqual(r, "file.txt")

        def on_error(err):
            assert False

        def on_completed():
            assert True

        res.subscribe(on_next, on_error, on_completed)

    def test_from_node_callback_mapper(self):
        res = reactivex.from_callback(lambda f, s, t, cb: cb(f, s, t), lambda r: r[0])(
            1, 2, 3
        )

        def on_next(r):
            self.assertEqual(r, 1)

        def on_error(err):
            assert False

        def on_completed():
            assert True

        res.subscribe(on_next, on_error, on_completed)



import unittest

import pytest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(ex):
    raise RxException(ex)


class TestRetry(unittest.TestCase):
    def test_retry_observable_basic(self):
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 1), on_next(150, 2), on_next(200, 3), on_completed(250)
        )
        results = scheduler.start(lambda: xs.pipe(ops.retry()))

        assert results.messages == [
            on_next(300, 1),
            on_next(350, 2),
            on_next(400, 3),
            on_completed(450),
        ]
        assert xs.subscriptions == [subscribe(200, 450)]

    def test_retry_observable_infinite(self):
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 1), on_next(150, 2), on_next(200, 3)
        )
        results = scheduler.start(lambda: xs.pipe(ops.retry()))

        assert results.messages == [on_next(300, 1), on_next(350, 2), on_next(400, 3)]
        assert xs.subscriptions == [subscribe(200, 1000)]

    def test_retry_observable_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 1), on_next(150, 2), on_next(200, 3), on_error(250, ex)
        )
        results = scheduler.start(lambda: xs.pipe(ops.retry()), disposed=1100)
        assert results.messages == [
            on_next(300, 1),
            on_next(350, 2),
            on_next(400, 3),
            on_next(550, 1),
            on_next(600, 2),
            on_next(650, 3),
            on_next(800, 1),
            on_next(850, 2),
            on_next(900, 3),
            on_next(1050, 1),
        ]
        assert xs.subscriptions == [
            subscribe(200, 450),
            subscribe(450, 700),
            subscribe(700, 950),
            subscribe(950, 1100),
        ]

    def test_retry_observable_throws(self):
        scheduler1 = TestScheduler()
        xs = reactivex.return_value(1).pipe(ops.retry())
        xs.subscribe(lambda x: _raise("ex"), scheduler=scheduler1)

        with pytest.raises(RxException):
            scheduler1.start()

        scheduler2 = TestScheduler()
        ys = reactivex.throw("ex").pipe(ops.retry())
        d = ys.subscribe(on_error=lambda ex: _raise("ex"), scheduler=scheduler2)

        scheduler2.schedule_absolute(210, lambda sc, st: d.dispose())
        scheduler2.start()

        scheduler3 = TestScheduler()
        zs = reactivex.return_value(1).pipe(ops.retry())
        zs.subscribe(on_completed=lambda: _raise("ex"), scheduler=scheduler3)

        with pytest.raises(RxException):
            scheduler3.start()

    def test_retry_observable_retry_count_basic(self):
        scheduler = TestScheduler()
        ex = "ex"
        xs = scheduler.create_cold_observable(
            on_next(5, 1), on_next(10, 2), on_next(15, 3), on_error(20, ex)
        )
        results = scheduler.start(lambda: xs.pipe(ops.retry(3)))

        assert results.messages == [
            on_next(205, 1),
            on_next(210, 2),
            on_next(215, 3),
            on_next(225, 1),
            on_next(230, 2),
            on_next(235, 3),
            on_next(245, 1),
            on_next(250, 2),
            on_next(255, 3),
            on_error(260, ex),
        ]
        assert xs.subscriptions == [
            subscribe(200, 220),
            subscribe(220, 240),
            subscribe(240, 260),
        ]

    def test_retry_observable_retry_count_dispose(self):
        scheduler = TestScheduler()
        ex = "ex"
        xs = scheduler.create_cold_observable(
            on_next(5, 1), on_next(10, 2), on_next(15, 3), on_error(20, ex)
        )
        results = scheduler.start(lambda: xs.pipe(ops.retry(3)), disposed=231)
        assert results.messages == [
            on_next(205, 1),
            on_next(210, 2),
            on_next(215, 3),
            on_next(225, 1),
            on_next(230, 2),
        ]
        assert xs.subscriptions == [subscribe(200, 220), subscribe(220, 231)]

    def test_retry_observable_retry_count_dispose_ii(self):
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 1), on_next(150, 2), on_next(200, 3)
        )
        results = scheduler.start(lambda: xs.pipe(ops.retry(3)))

        assert results.messages == [on_next(300, 1), on_next(350, 2), on_next(400, 3)]
        assert xs.subscriptions == [subscribe(200, 1000)]

    def test_retry_observable_retry_count_dispose_iii(self):
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 1), on_next(150, 2), on_next(200, 3), on_completed(250)
        )
        results = scheduler.start(lambda: xs.pipe(ops.retry(3)))

        assert results.messages == [
            on_next(300, 1),
            on_next(350, 2),
            on_next(400, 3),
            on_completed(450),
        ]
        assert xs.subscriptions == [subscribe(200, 450)]

    def test_retry_observable_retry_count_throws(self):
        scheduler1 = TestScheduler()
        xs = reactivex.return_value(1).pipe(ops.retry(3))
        xs.subscribe(lambda x: _raise("ex"), scheduler=scheduler1)

        self.assertRaises(RxException, scheduler1.start)

        scheduler2 = TestScheduler()
        ys = reactivex.throw("ex").pipe(ops.retry(100))
        d = ys.subscribe(on_error=lambda ex: _raise("ex"), scheduler=scheduler2)

        def dispose(_, __):
            d.dispose()

        scheduler2.schedule_absolute(0, dispose)
        scheduler2.start()

        scheduler3 = TestScheduler()
        zs = reactivex.return_value(1).pipe(ops.retry(100))
        zs.subscribe(on_completed=lambda: _raise("ex"), scheduler=scheduler3)

        with pytest.raises(RxException):
            scheduler3.start()

        xss = reactivex.create(lambda o: _raise("ex")).pipe(ops.retry(100))
        with pytest.raises(Exception):
            xss.subscribe()


if __name__ == "__main__":
    unittest.main()



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestBufferWithCount(unittest.TestCase):
    def test_buffer_with_time_basic(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(281, 4),
            on_next(321, 5),
            on_next(351, 6),
            on_next(381, 7),
            on_next(421, 8),
            on_next(471, 9),
            on_completed(600),
        )

        def create():
            return xs.pipe(
                ops.buffer_with_time(100, 70),
                ops.map(lambda x: ",".join([str(a) for a in x])),
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(300, "2,3,4"),
            on_next(370, "4,5,6"),
            on_next(440, "6,7,8"),
            on_next(510, "8,9"),
            on_next(580, ""),
            on_next(600, ""),
            on_completed(600),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]

    def test_buffer_with_time_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(280, 4),
            on_next(320, 5),
            on_next(350, 6),
            on_next(380, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_error(600, ex),
        )

        def create():
            return xs.pipe(
                ops.buffer_with_time(100, 70),
                ops.map(lambda x: ",".join([str(a) for a in x])),
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(300, "2,3,4"),
            on_next(370, "4,5,6"),
            on_next(440, "6,7,8"),
            on_next(510, "8,9"),
            on_next(580, ""),
            on_error(600, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]

    def test_buffer_with_time_disposed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(281, 4),
            on_next(321, 5),
            on_next(351, 6),
            on_next(381, 7),
            on_next(421, 8),
            on_next(471, 9),
            on_completed(600),
        )

        def create():
            return xs.pipe(
                ops.buffer_with_time(100, 70),
                ops.map(lambda x: ",".join([str(a) for a in x])),
            )

        results = scheduler.start(create, disposed=370)
        assert results.messages == [on_next(300, "2,3,4")]
        assert xs.subscriptions == [subscribe(200, 370)]

    def test_buffer_with_time_basic_same(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(280, 4),
            on_next(320, 5),
            on_next(350, 6),
            on_next(380, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_completed(600),
        )

        def create():
            return xs.pipe(
                ops.buffer_with_time(100),
                ops.map(lambda x: ",".join([str(a) for a in x])),
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(300, "2,3,4"),
            on_next(400, "5,6,7"),
            on_next(500, "8,9"),
            on_next(600, ""),
            on_completed(600),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]



import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestOnErrorResumeNext(unittest.TestCase):
    def test_on_error_resume_next_no_errors(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_next(220, 3), on_completed(230)]

        msgs2 = [on_next(240, 4), on_completed(250)]

        o1 = scheduler.create_hot_observable(msgs1)
        o2 = scheduler.create_hot_observable(msgs2)

        def create():
            return o1.pipe(ops.on_error_resume_next(o2))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(210, 2),
            on_next(220, 3),
            on_next(240, 4),
            on_completed(250),
        ]

    def test_on_error_resume_next_error(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_next(220, 3), on_error(230, "ex")]

        msgs2 = [on_next(240, 4), on_completed(250)]

        o1 = scheduler.create_hot_observable(msgs1)
        o2 = scheduler.create_hot_observable(msgs2)

        def create():
            return o1.pipe(ops.on_error_resume_next(o2))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(210, 2),
            on_next(220, 3),
            on_next(240, 4),
            on_completed(250),
        ]

    def test_on_error_resume_next_error_multiple(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_error(220, "ex")]

        msgs2 = [on_next(230, 4), on_error(240, "ex")]

        msgs3 = [on_completed(250)]

        o1 = scheduler.create_hot_observable(msgs1)
        o2 = scheduler.create_hot_observable(msgs2)
        o3 = scheduler.create_hot_observable(msgs3)

        def create():
            return reactivex.on_error_resume_next(o1, o2, o3)

        results = scheduler.start(create)

        assert results.messages == [on_next(210, 2), on_next(230, 4), on_completed(250)]

    def test_on_error_resume_next_empty_return_throw_and_more(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(205)]
        msgs2 = [on_next(215, 2), on_completed(220)]
        msgs3 = [on_next(225, 3), on_next(230, 4), on_completed(235)]
        msgs4 = [on_error(240, "ex")]
        msgs5 = [on_next(245, 5), on_completed(250)]

        o1 = scheduler.create_hot_observable(msgs1)
        o2 = scheduler.create_hot_observable(msgs2)
        o3 = scheduler.create_hot_observable(msgs3)
        o4 = scheduler.create_hot_observable(msgs4)
        o5 = scheduler.create_hot_observable(msgs5)

        def create():
            return reactivex.on_error_resume_next(o1, o2, o3, o4, o5)

        results = scheduler.start(create)

        assert results.messages == [
            on_next(215, 2),
            on_next(225, 3),
            on_next(230, 4),
            on_next(245, 5),
            on_completed(250),
        ]

    def test_on_error_resume_next_empty_return_throw_and_more_ii(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_completed(220)]
        msgs2 = [on_error(230, ex)]
        o1 = scheduler.create_hot_observable(msgs1)
        o2 = scheduler.create_hot_observable(msgs2)

        def create():
            return o1.pipe(ops.on_error_resume_next(o2))

        results = scheduler.start(create)

        assert results.messages == [on_next(210, 2), on_completed(230)]

    def test_on_error_resume_next_single_source_throws(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_error(230, ex)]
        o1 = scheduler.create_hot_observable(msgs1)

        def create():
            return reactivex.on_error_resume_next(o1)

        results = scheduler.start(create)

        assert results.messages == [on_completed(230)]

    def test_on_error_resume_next_end_with_never(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_completed(220)]
        o1 = scheduler.create_hot_observable(msgs1)
        o2 = reactivex.never()

        def create():
            return reactivex.on_error_resume_next(o1, o2)

        results = scheduler.start(create)

        assert results.messages == [on_next(210, 2)]

    def test_on_error_resume_next_start_with_never(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_completed(220)]
        o1 = reactivex.never()
        o2 = scheduler.create_hot_observable(msgs1)

        def create():
            return reactivex.on_error_resume_next(o1, o2)

        results = scheduler.start(create)

        assert results.messages == []

    def test_on_error_resume_next_start_with_factory(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_next(220, 3), on_error(230, "ex")]
        o1 = scheduler.create_hot_observable(msgs1)

        def factory(ex: Exception):
            assert str(ex) == "ex"
            msgs2 = [on_next(240, 4), on_completed(250)]
            o2 = scheduler.create_hot_observable(msgs2)
            return o2

        def create():
            return reactivex.on_error_resume_next(o1, factory)

        results = scheduler.start(create)

        assert results.messages == [
            on_next(210, 2),
            on_next(220, 3),
            on_next(240, 4),
            on_completed(250),
        ]



import os
import threading
import unittest
from datetime import timedelta
from time import sleep

import pytest

from reactivex.disposable import Disposable
from reactivex.internal.basic import default_now
from reactivex.internal.constants import DELTA_ZERO
from reactivex.internal.exceptions import WouldBlockException
from reactivex.scheduler import ImmediateScheduler

CI = os.getenv("CI") is not None


class TestImmediateScheduler(unittest.TestCase):
    def test_immediate_singleton(self):
        scheduler = [ImmediateScheduler(), ImmediateScheduler.singleton()]
        assert scheduler[0] is scheduler[1]

        gate = [threading.Semaphore(0), threading.Semaphore(0)]
        scheduler = [None, None]

        def run(idx):
            scheduler[idx] = ImmediateScheduler()
            gate[idx].release()

        for idx in (0, 1):
            threading.Thread(target=run, args=(idx,)).start()
            gate[idx].acquire()

        assert scheduler[0] is not None
        assert scheduler[1] is not None
        assert scheduler[0] is scheduler[1]

    def test_immediate_extend(self):
        class MyScheduler(ImmediateScheduler):
            pass

        scheduler = [
            MyScheduler(),
            MyScheduler.singleton(),
            ImmediateScheduler.singleton(),
        ]
        assert scheduler[0] is scheduler[1]
        assert scheduler[0] is not scheduler[2]

    @pytest.mark.skipif(CI, reason="Flaky test in GitHub Actions")
    def test_immediate_now(self):
        scheduler = ImmediateScheduler()
        diff = scheduler.now - default_now()
        assert abs(diff) <= timedelta(milliseconds=1)

    @pytest.mark.skipif(CI, reason="Flaky test in GitHub Actions")
    def test_immediate_now_units(self):
        scheduler = ImmediateScheduler()
        diff = scheduler.now
        sleep(1.1)
        diff = scheduler.now - diff
        assert timedelta(milliseconds=1000) < diff < timedelta(milliseconds=1300)

    def test_immediate_scheduleaction(self):
        scheduler = ImmediateScheduler()
        ran = False

        def action(scheduler, state=None):
            nonlocal ran
            ran = True

        scheduler.schedule(action)
        assert ran

    def test_immediate_schedule_action_error(self):
        scheduler = ImmediateScheduler()

        class MyException(Exception):
            pass

        def action(scheduler, state=None):
            raise MyException()

        with pytest.raises(MyException):
            return scheduler.schedule(action)

    def test_immediate_schedule_action_due_error(self):
        scheduler = ImmediateScheduler()
        ran = False

        def action(scheduler, state=None):
            nonlocal ran
            ran = True

        with pytest.raises(WouldBlockException):
            scheduler.schedule_relative(0.1, action)

        assert ran is False

    def test_immediate_simple1(self):
        scheduler = ImmediateScheduler()
        xx = 0

        def action(scheduler, state=None):
            nonlocal xx
            xx = state
            return Disposable()

        scheduler.schedule(action, 42)
        assert xx == 42

    def test_immediate_simple2(self):
        scheduler = ImmediateScheduler()
        xx = 0

        def action(scheduler, state=None):
            nonlocal xx
            xx = state
            return Disposable()

        scheduler.schedule_absolute(default_now(), action, 42)
        assert xx == 42

    def test_immediate_simple3(self):
        scheduler = ImmediateScheduler()
        xx = 0

        def action(scheduler, state=None):
            nonlocal xx
            xx = state
            return Disposable()

        scheduler.schedule_relative(DELTA_ZERO, action, 42)
        assert xx == 42

    def test_immediate_recursive1(self):
        scheduler = ImmediateScheduler()
        xx = 0
        yy = 0

        def action(scheduler, state=None):
            nonlocal xx
            xx = state

            def inner_action(scheduler, state=None):
                nonlocal yy
                yy = state
                return Disposable()

            return scheduler.schedule(inner_action, 43)

        scheduler.schedule(action, 42)
        assert xx == 42
        assert yy == 43

    def test_immediate_recursive2(self):
        scheduler = ImmediateScheduler()
        xx = 0
        yy = 0

        def action(scheduler, state=None):
            nonlocal xx
            xx = state

            def inner_action(scheduler, state=None):
                nonlocal yy
                yy = state
                return Disposable()

            return scheduler.schedule_absolute(default_now(), inner_action, 43)

        scheduler.schedule_absolute(default_now(), action, 42)

        assert xx == 42
        assert yy == 43

    def test_immediate_recursive3(self):
        scheduler = ImmediateScheduler()
        xx = 0
        yy = 0

        def action(scheduler, state=None):
            nonlocal xx
            xx = state

            def inner_action(scheduler, state):
                nonlocal yy
                yy = state
                return Disposable()

            return scheduler.schedule_relative(DELTA_ZERO, inner_action, 43)

        scheduler.schedule_relative(DELTA_ZERO, action, 42)

        assert xx == 42
        assert yy == 43



import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestWindowWithTime(unittest.TestCase):
    def test_window_time_basic(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(270, 4),
            on_next(320, 5),
            on_next(360, 6),
            on_next(390, 7),
            on_next(410, 8),
            on_next(460, 9),
            on_next(470, 10),
            on_completed(490),
        )

        def create():
            def mapper(ys, i):
                def proj(y):
                    return "%s %s" % (i, y)

                return ys.pipe(
                    ops.map(proj), ops.concat(reactivex.return_value("%s end" % i))
                )

            return xs.pipe(
                ops.window_with_time(100), ops.map_indexed(mapper), ops.merge_all()
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(210, "0 2"),
            on_next(240, "0 3"),
            on_next(270, "0 4"),
            on_next(300, "0 end"),
            on_next(320, "1 5"),
            on_next(360, "1 6"),
            on_next(390, "1 7"),
            on_next(400, "1 end"),
            on_next(410, "2 8"),
            on_next(460, "2 9"),
            on_next(470, "2 10"),
            on_next(490, "2 end"),
            on_completed(490),
        ]
        assert xs.subscriptions == [subscribe(200, 490)]

    def test_window_time_basic_both(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(270, 4),
            on_next(320, 5),
            on_next(360, 6),
            on_next(390, 7),
            on_next(410, 8),
            on_next(460, 9),
            on_next(470, 10),
            on_completed(490),
        )

        def create():
            def mapper(ys, i):
                def proj(y):
                    return "%s %s" % (i, y)

                return ys.pipe(
                    ops.map(proj), ops.concat(reactivex.return_value("%s end" % i))
                )

            return xs.pipe(
                ops.window_with_time(100, 50), ops.map_indexed(mapper), ops.merge_all()
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(210, "0 2"),
            on_next(240, "0 3"),
            on_next(270, "0 4"),
            on_next(270, "1 4"),
            on_next(300, "0 end"),
            on_next(320, "1 5"),
            on_next(320, "2 5"),
            on_next(350, "1 end"),
            on_next(360, "2 6"),
            on_next(360, "3 6"),
            on_next(390, "2 7"),
            on_next(390, "3 7"),
            on_next(400, "2 end"),
            on_next(410, "3 8"),
            on_next(410, "4 8"),
            on_next(450, "3 end"),
            on_next(460, "4 9"),
            on_next(460, "5 9"),
            on_next(470, "4 10"),
            on_next(470, "5 10"),
            on_next(490, "4 end"),
            on_next(490, "5 end"),
            on_completed(490),
        ]
        assert xs.subscriptions == [subscribe(200, 490)]

    def test_window_with_time_basic(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(280, 4),
            on_next(320, 5),
            on_next(350, 6),
            on_next(380, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_completed(600),
        )

        def create():
            def mapper(w, i):
                return w.pipe(ops.map(lambda x: "%s %s" % (i, x)))

            return xs.pipe(
                ops.window_with_time(100, 70), ops.map_indexed(mapper), ops.merge_all()
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(210, "0 2"),
            on_next(240, "0 3"),
            on_next(280, "0 4"),
            on_next(280, "1 4"),
            on_next(320, "1 5"),
            on_next(350, "1 6"),
            on_next(350, "2 6"),
            on_next(380, "2 7"),
            on_next(420, "2 8"),
            on_next(420, "3 8"),
            on_next(470, "3 9"),
            on_completed(600),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]

    def test_window_with_time_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(280, 4),
            on_next(320, 5),
            on_next(350, 6),
            on_next(380, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_error(600, ex),
        )

        def create():
            def mapper(w, i):
                return w.pipe(ops.map(lambda x: "%s %s" % (i, x)))

            return xs.pipe(
                ops.window_with_time(100, 70), ops.map_indexed(mapper), ops.merge_all()
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(210, "0 2"),
            on_next(240, "0 3"),
            on_next(280, "0 4"),
            on_next(280, "1 4"),
            on_next(320, "1 5"),
            on_next(350, "1 6"),
            on_next(350, "2 6"),
            on_next(380, "2 7"),
            on_next(420, "2 8"),
            on_next(420, "3 8"),
            on_next(470, "3 9"),
            on_error(600, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]

    def test_Window_with_time_disposed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(280, 4),
            on_next(320, 5),
            on_next(350, 6),
            on_next(380, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_completed(600),
        )

        def create():
            def mapper(w, i):
                return w.pipe(ops.map(lambda x: "%s %s" % (i, x)))

            return xs.pipe(
                ops.window_with_time(100, 70), ops.map_indexed(mapper), ops.merge_all()
            )

        results = scheduler.start(create, disposed=370)
        assert results.messages == [
            on_next(210, "0 2"),
            on_next(240, "0 3"),
            on_next(280, "0 4"),
            on_next(280, "1 4"),
            on_next(320, "1 5"),
            on_next(350, "1 6"),
            on_next(350, "2 6"),
        ]
        assert xs.subscriptions == [subscribe(200, 370)]

    def test_window_with_time_basic_same(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(280, 4),
            on_next(320, 5),
            on_next(350, 6),
            on_next(380, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_completed(600),
        )

        def create():
            def mapper(w, i):
                return w.pipe(ops.map(lambda x: "%s %s" % (i, x)))

            return xs.pipe(
                ops.window_with_time(100), ops.map_indexed(mapper), ops.merge_all()
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(210, "0 2"),
            on_next(240, "0 3"),
            on_next(280, "0 4"),
            on_next(320, "1 5"),
            on_next(350, "1 6"),
            on_next(380, "1 7"),
            on_next(420, "2 8"),
            on_next(470, "2 9"),
            on_completed(600),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]



"""
RxPY example running a Tornado server doing search queries against Wikipedia to
populate the autocomplete dropdown in the web UI. Start using
`python autocomplete.py` and navigate your web browser to http://localhost:8080

Uses the RxPY IOLoopScheduler.
"""

import os
from asyncio import Future
from typing import Any, Dict, Union

from tornado import ioloop
from tornado.escape import json_decode
from tornado.httpclient import AsyncHTTPClient, HTTPResponse
from tornado.httputil import url_concat
from tornado.web import Application, RequestHandler, StaticFileHandler, url
from tornado.websocket import WebSocketHandler

from reactivex import operators as ops
from reactivex.scheduler.eventloop import IOLoopScheduler
from reactivex.subject import Subject

scheduler = IOLoopScheduler(ioloop.IOLoop.current())


def search_wikipedia(term: str) -> Future[HTTPResponse]:
    """Search Wikipedia for a given term"""
    url = "http://en.wikipedia.org/w/api.php"

    params: Dict[str, str] = {"action": "opensearch", "search": term, "format": "json"}
    # Must set a user agent for non-browser requests to Wikipedia
    user_agent = (
        "RxPY/3.0 (https://github.com/dbrattli/RxPY; dag@brattli.net) Tornado/4.0.1"
    )

    url = url_concat(url, params)

    http_client = AsyncHTTPClient()
    return http_client.fetch(url, method="GET", user_agent=user_agent)


class WSHandler(WebSocketHandler):
    def open(self, *args: Any):
        print("WebSocket opened")

        # A Subject is both an observable and observer, so we can both subscribe
        # to it and also feed (send) it with new values
        self.subject: Subject[Dict[str, str]] = Subject()

        # Get all distinct key up events from the input and only fire if long enough and distinct
        searcher = self.subject.pipe(
            ops.map(lambda x: x["term"]),
            ops.filter(
                lambda text: len(text) > 2
            ),  # Only if the text is longer than 2 characters
            ops.debounce(0.750),  # Pause for 750ms
            ops.distinct_until_changed(),  # Only if the value has changed
            ops.flat_map_latest(search_wikipedia),
        )

        def send_response(x: HTTPResponse) -> None:
            self.write_message(x.body)

        def on_error(ex: Exception) -> None:
            print(ex)

        searcher.subscribe(
            on_next=send_response, on_error=on_error, scheduler=scheduler
        )

    def on_message(self, message: Union[bytes, str]):
        obj = json_decode(message)
        self.subject.on_next(obj)

    def on_close(self):
        print("WebSocket closed")


class MainHandler(RequestHandler):
    def get(self):
        self.render("index.html")


def main():
    port = os.environ.get("PORT", 8080)
    app = Application(
        [
            url(r"/", MainHandler),
            (r"/ws", WSHandler),
            (r"/static/(.*)", StaticFileHandler, {"path": "."}),
        ]
    )
    print("Starting server at port: %s" % port)
    app.listen(port)
    ioloop.IOLoop.current().start()


if __name__ == "__main__":
    main()



import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestFlatMap(unittest.TestCase):
    def test_flat_map_then_complete_complete(self):
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 4),
            on_next(200, 2),
            on_next(300, 3),
            on_next(400, 1),
            on_completed(500),
        )
        ys = scheduler.create_cold_observable(
            on_next(50, "foo"),
            on_next(100, "bar"),
            on_next(150, "baz"),
            on_next(200, "qux"),
            on_completed(250),
        )

        def factory():
            return xs.pipe(ops.flat_map(ys))

        results = scheduler.start(factory)

        assert results.messages == [
            on_next(350, "foo"),
            on_next(400, "bar"),
            on_next(450, "baz"),
            on_next(450, "foo"),
            on_next(500, "qux"),
            on_next(500, "bar"),
            on_next(550, "baz"),
            on_next(550, "foo"),
            on_next(600, "qux"),
            on_next(600, "bar"),
            on_next(650, "baz"),
            on_next(650, "foo"),
            on_next(700, "qux"),
            on_next(700, "bar"),
            on_next(750, "baz"),
            on_next(800, "qux"),
            on_completed(850),
        ]
        assert xs.subscriptions == [subscribe(200, 700)]
        assert ys.subscriptions == [
            subscribe(300, 550),
            subscribe(400, 650),
            subscribe(500, 750),
            subscribe(600, 850),
        ]

    def test_flat_map_then_complete_complete_2(self):
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 4),
            on_next(200, 2),
            on_next(300, 3),
            on_next(400, 1),
            on_completed(700),
        )
        ys = scheduler.create_cold_observable(
            on_next(50, "foo"),
            on_next(100, "bar"),
            on_next(150, "baz"),
            on_next(200, "qux"),
            on_completed(250),
        )

        def factory():
            return xs.pipe(ops.flat_map(ys))

        results = scheduler.start(factory)

        assert results.messages == [
            on_next(350, "foo"),
            on_next(400, "bar"),
            on_next(450, "baz"),
            on_next(450, "foo"),
            on_next(500, "qux"),
            on_next(500, "bar"),
            on_next(550, "baz"),
            on_next(550, "foo"),
            on_next(600, "qux"),
            on_next(600, "bar"),
            on_next(650, "baz"),
            on_next(650, "foo"),
            on_next(700, "qux"),
            on_next(700, "bar"),
            on_next(750, "baz"),
            on_next(800, "qux"),
            on_completed(900),
        ]
        assert xs.subscriptions == [subscribe(200, 900)]
        assert ys.subscriptions == [
            subscribe(300, 550),
            subscribe(400, 650),
            subscribe(500, 750),
            subscribe(600, 850),
        ]

    def test_flat_map_then_never_complete(self):
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 4),
            on_next(200, 2),
            on_next(300, 3),
            on_next(400, 1),
            on_next(500, 5),
            on_next(700, 0),
        )
        ys = scheduler.create_cold_observable(
            on_next(50, "foo"),
            on_next(100, "bar"),
            on_next(150, "baz"),
            on_next(200, "qux"),
            on_completed(250),
        )
        results = scheduler.start(lambda: xs.pipe(ops.flat_map(ys)))

        assert results.messages == [
            on_next(350, "foo"),
            on_next(400, "bar"),
            on_next(450, "baz"),
            on_next(450, "foo"),
            on_next(500, "qux"),
            on_next(500, "bar"),
            on_next(550, "baz"),
            on_next(550, "foo"),
            on_next(600, "qux"),
            on_next(600, "bar"),
            on_next(650, "baz"),
            on_next(650, "foo"),
            on_next(700, "qux"),
            on_next(700, "bar"),
            on_next(750, "baz"),
            on_next(750, "foo"),
            on_next(800, "qux"),
            on_next(800, "bar"),
            on_next(850, "baz"),
            on_next(900, "qux"),
            on_next(950, "foo"),
        ]
        assert xs.subscriptions == [subscribe(200, 1000)]
        assert ys.subscriptions == [
            subscribe(300, 550),
            subscribe(400, 650),
            subscribe(500, 750),
            subscribe(600, 850),
            subscribe(700, 950),
            subscribe(900, 1000),
        ]

    def test_flat_map_then_complete_never(self):
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 4),
            on_next(200, 2),
            on_next(300, 3),
            on_next(400, 1),
            on_completed(500),
        )
        ys = scheduler.create_cold_observable(
            on_next(50, "foo"),
            on_next(100, "bar"),
            on_next(150, "baz"),
            on_next(200, "qux"),
        )
        results = scheduler.start(lambda: xs.pipe(ops.flat_map(ys)))

        assert results.messages == [
            on_next(350, "foo"),
            on_next(400, "bar"),
            on_next(450, "baz"),
            on_next(450, "foo"),
            on_next(500, "qux"),
            on_next(500, "bar"),
            on_next(550, "baz"),
            on_next(550, "foo"),
            on_next(600, "qux"),
            on_next(600, "bar"),
            on_next(650, "baz"),
            on_next(650, "foo"),
            on_next(700, "qux"),
            on_next(700, "bar"),
            on_next(750, "baz"),
            on_next(800, "qux"),
        ]
        assert xs.subscriptions == [subscribe(200, 700)]
        assert ys.subscriptions == [
            subscribe(300, 1000),
            subscribe(400, 1000),
            subscribe(500, 1000),
            subscribe(600, 1000),
        ]

    def test_flat_map_then_complete_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 4),
            on_next(200, 2),
            on_next(300, 3),
            on_next(400, 1),
            on_completed(500),
        )
        ys = scheduler.create_cold_observable(
            on_next(50, "foo"),
            on_next(100, "bar"),
            on_next(150, "baz"),
            on_next(200, "qux"),
            on_error(300, ex),
        )
        results = scheduler.start(lambda: xs.pipe(ops.flat_map(ys)))

        assert results.messages == [
            on_next(350, "foo"),
            on_next(400, "bar"),
            on_next(450, "baz"),
            on_next(450, "foo"),
            on_next(500, "qux"),
            on_next(500, "bar"),
            on_next(550, "baz"),
            on_next(550, "foo"),
            on_error(600, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]
        assert ys.subscriptions == [
            subscribe(300, 600),
            subscribe(400, 600),
            subscribe(500, 600),
            subscribe(600, 600),
        ]

    def test_flat_map_then_error_complete(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 4),
            on_next(200, 2),
            on_next(300, 3),
            on_next(400, 1),
            on_error(500, ex),
        )
        ys = scheduler.create_cold_observable(
            on_next(50, "foo"),
            on_next(100, "bar"),
            on_next(150, "baz"),
            on_next(200, "qux"),
            on_completed(250),
        )
        results = scheduler.start(lambda: xs.pipe(ops.flat_map(ys)))

        assert results.messages == [
            on_next(350, "foo"),
            on_next(400, "bar"),
            on_next(450, "baz"),
            on_next(450, "foo"),
            on_next(500, "qux"),
            on_next(500, "bar"),
            on_next(550, "baz"),
            on_next(550, "foo"),
            on_next(600, "qux"),
            on_next(600, "bar"),
            on_next(650, "baz"),
            on_next(650, "foo"),
            on_error(700, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 700)]
        assert ys.subscriptions == [
            subscribe(300, 550),
            subscribe(400, 650),
            subscribe(500, 700),
            subscribe(600, 700),
        ]

    def test_flat_map_then_error_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_cold_observable(
            on_next(100, 4),
            on_next(200, 2),
            on_next(300, 3),
            on_next(400, 1),
            on_error(500, ex),
        )
        ys = scheduler.create_cold_observable(
            on_next(50, "foo"),
            on_next(100, "bar"),
            on_next(150, "baz"),
            on_next(200, "qux"),
            on_error(250, ex),
        )
        results = scheduler.start(lambda: xs.pipe(ops.flat_map(ys)))

        assert results.messages == [
            on_next(350, "foo"),
            on_next(400, "bar"),
            on_next(450, "baz"),
            on_next(450, "foo"),
            on_next(500, "qux"),
            on_next(500, "bar"),
            on_error(550, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 550)]
        assert ys.subscriptions == [
            subscribe(300, 550),
            subscribe(400, 550),
            subscribe(500, 550),
        ]

    def test_flat_map_complete(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(5, scheduler.create_cold_observable(on_error(1, "ex1"))),
            on_next(105, scheduler.create_cold_observable(on_error(1, "ex2"))),
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 102),
                    on_next(90, 103),
                    on_next(110, 104),
                    on_next(190, 105),
                    on_next(440, 106),
                    on_completed(460),
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(
                    on_next(180, 202), on_next(190, 203), on_completed(205)
                ),
            ),
            on_next(
                550,
                scheduler.create_cold_observable(
                    on_next(10, 301),
                    on_next(50, 302),
                    on_next(70, 303),
                    on_next(260, 304),
                    on_next(310, 305),
                    on_completed(410),
                ),
            ),
            on_next(750, scheduler.create_cold_observable(on_completed(40))),
            on_next(
                850,
                scheduler.create_cold_observable(
                    on_next(80, 401), on_next(90, 402), on_completed(100)
                ),
            ),
            on_completed(900),
        )

        def factory():
            return xs.pipe(ops.flat_map(lambda x: x))

        results = scheduler.start(factory)

        assert results.messages == [
            on_next(310, 102),
            on_next(390, 103),
            on_next(410, 104),
            on_next(490, 105),
            on_next(560, 301),
            on_next(580, 202),
            on_next(590, 203),
            on_next(600, 302),
            on_next(620, 303),
            on_next(740, 106),
            on_next(810, 304),
            on_next(860, 305),
            on_next(930, 401),
            on_next(940, 402),
            on_completed(960),
        ]
        assert xs.subscriptions == [subscribe(200, 900)]
        assert xs.messages[2].value.value.subscriptions == [subscribe(300, 760)]
        assert xs.messages[3].value.value.subscriptions == [subscribe(400, 605)]
        assert xs.messages[4].value.value.subscriptions == [subscribe(550, 960)]
        assert xs.messages[5].value.value.subscriptions == [subscribe(750, 790)]
        assert xs.messages[6].value.value.subscriptions == [subscribe(850, 950)]

    def test_flat_map_complete_inner_not_complete(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(5, scheduler.create_cold_observable(on_error(1, "ex1"))),
            on_next(105, scheduler.create_cold_observable(on_error(1, "ex2"))),
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 102),
                    on_next(90, 103),
                    on_next(110, 104),
                    on_next(190, 105),
                    on_next(440, 106),
                    on_completed(460),
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(on_next(180, 202), on_next(190, 203)),
            ),
            on_next(
                550,
                scheduler.create_cold_observable(
                    on_next(10, 301),
                    on_next(50, 302),
                    on_next(70, 303),
                    on_next(260, 304),
                    on_next(310, 305),
                    on_completed(410),
                ),
            ),
            on_next(750, scheduler.create_cold_observable(on_completed(40))),
            on_next(
                850,
                scheduler.create_cold_observable(
                    on_next(80, 401), on_next(90, 402), on_completed(100)
                ),
            ),
            on_completed(900),
        )

        def factory():
            return xs.pipe(ops.flat_map(lambda x: x))

        results = scheduler.start(factory)

        assert results.messages == [
            on_next(310, 102),
            on_next(390, 103),
            on_next(410, 104),
            on_next(490, 105),
            on_next(560, 301),
            on_next(580, 202),
            on_next(590, 203),
            on_next(600, 302),
            on_next(620, 303),
            on_next(740, 106),
            on_next(810, 304),
            on_next(860, 305),
            on_next(930, 401),
            on_next(940, 402),
        ]
        assert xs.subscriptions == [subscribe(200, 900)]
        assert xs.messages[2].value.value.subscriptions == [subscribe(300, 760)]
        assert xs.messages[3].value.value.subscriptions == [subscribe(400, 1000)]
        assert xs.messages[4].value.value.subscriptions == [subscribe(550, 960)]
        assert xs.messages[5].value.value.subscriptions == [subscribe(750, 790)]
        assert xs.messages[6].value.value.subscriptions == [subscribe(850, 950)]

    def test_flat_map_complete_outer_not_complete(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(5, scheduler.create_cold_observable(on_error(1, "ex1"))),
            on_next(105, scheduler.create_cold_observable(on_error(1, "ex2"))),
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 102),
                    on_next(90, 103),
                    on_next(110, 104),
                    on_next(190, 105),
                    on_next(440, 106),
                    on_completed(460),
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(
                    on_next(180, 202), on_next(190, 203), on_completed(205)
                ),
            ),
            on_next(
                550,
                scheduler.create_cold_observable(
                    on_next(10, 301),
                    on_next(50, 302),
                    on_next(70, 303),
                    on_next(260, 304),
                    on_next(310, 305),
                    on_completed(410),
                ),
            ),
            on_next(750, scheduler.create_cold_observable(on_completed(40))),
            on_next(
                850,
                scheduler.create_cold_observable(
                    on_next(80, 401), on_next(90, 402), on_completed(100)
                ),
            ),
        )

        def factory():
            return xs.pipe(ops.flat_map(lambda x: x))

        results = scheduler.start(factory)

        assert results.messages == [
            on_next(310, 102),
            on_next(390, 103),
            on_next(410, 104),
            on_next(490, 105),
            on_next(560, 301),
            on_next(580, 202),
            on_next(590, 203),
            on_next(600, 302),
            on_next(620, 303),
            on_next(740, 106),
            on_next(810, 304),
            on_next(860, 305),
            on_next(930, 401),
            on_next(940, 402),
        ]
        assert xs.subscriptions == [subscribe(200, 1000)]
        assert xs.messages[2].value.value.subscriptions == [subscribe(300, 760)]
        assert xs.messages[3].value.value.subscriptions == [subscribe(400, 605)]
        assert xs.messages[4].value.value.subscriptions == [subscribe(550, 960)]
        assert xs.messages[5].value.value.subscriptions == [subscribe(750, 790)]
        assert xs.messages[6].value.value.subscriptions == [subscribe(850, 950)]

    def test_flat_map_error_outer(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(5, scheduler.create_cold_observable(on_error(1, "ex1"))),
            on_next(105, scheduler.create_cold_observable(on_error(1, "ex2"))),
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 102),
                    on_next(90, 103),
                    on_next(110, 104),
                    on_next(190, 105),
                    on_next(440, 106),
                    on_completed(460),
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(
                    on_next(180, 202), on_next(190, 203), on_completed(205)
                ),
            ),
            on_next(
                550,
                scheduler.create_cold_observable(
                    on_next(10, 301),
                    on_next(50, 302),
                    on_next(70, 303),
                    on_next(260, 304),
                    on_next(310, 305),
                    on_completed(410),
                ),
            ),
            on_next(750, scheduler.create_cold_observable(on_completed(40))),
            on_next(
                850,
                scheduler.create_cold_observable(
                    on_next(80, 401), on_next(90, 402), on_completed(100)
                ),
            ),
            on_error(900, ex),
        )

        def factory():
            return xs.pipe(ops.flat_map(lambda x: x))

        results = scheduler.start(factory)

        assert results.messages == [
            on_next(310, 102),
            on_next(390, 103),
            on_next(410, 104),
            on_next(490, 105),
            on_next(560, 301),
            on_next(580, 202),
            on_next(590, 203),
            on_next(600, 302),
            on_next(620, 303),
            on_next(740, 106),
            on_next(810, 304),
            on_next(860, 305),
            on_error(900, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 900)]
        assert xs.messages[2].value.value.subscriptions == [subscribe(300, 760)]
        assert xs.messages[3].value.value.subscriptions == [subscribe(400, 605)]
        assert xs.messages[4].value.value.subscriptions == [subscribe(550, 900)]
        assert xs.messages[5].value.value.subscriptions == [subscribe(750, 790)]
        assert xs.messages[6].value.value.subscriptions == [subscribe(850, 900)]

    def test_flat_map_error_inner(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(5, scheduler.create_cold_observable(on_error(1, "ex1"))),
            on_next(105, scheduler.create_cold_observable(on_error(1, "ex2"))),
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 102),
                    on_next(90, 103),
                    on_next(110, 104),
                    on_next(190, 105),
                    on_next(440, 106),
                    on_error(460, ex),
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(
                    on_next(180, 202), on_next(190, 203), on_completed(205)
                ),
            ),
            on_next(
                550,
                scheduler.create_cold_observable(
                    on_next(10, 301),
                    on_next(50, 302),
                    on_next(70, 303),
                    on_next(260, 304),
                    on_next(310, 305),
                    on_completed(410),
                ),
            ),
            on_next(750, scheduler.create_cold_observable(on_completed(40))),
            on_next(
                850,
                scheduler.create_cold_observable(
                    on_next(80, 401), on_next(90, 402), on_completed(100)
                ),
            ),
            on_completed(900),
        )

        def factory():
            return xs.pipe(ops.flat_map(lambda x: x))

        results = scheduler.start(factory)

        assert results.messages == [
            on_next(310, 102),
            on_next(390, 103),
            on_next(410, 104),
            on_next(490, 105),
            on_next(560, 301),
            on_next(580, 202),
            on_next(590, 203),
            on_next(600, 302),
            on_next(620, 303),
            on_next(740, 106),
            on_error(760, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 760)]
        assert xs.messages[2].value.value.subscriptions == [subscribe(300, 760)]
        assert xs.messages[3].value.value.subscriptions == [subscribe(400, 605)]
        assert xs.messages[4].value.value.subscriptions == [subscribe(550, 760)]
        assert xs.messages[5].value.value.subscriptions == [subscribe(750, 760)]
        assert xs.messages[6].value.value.subscriptions == []

    def test_flat_map_dispose(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(5, scheduler.create_cold_observable(on_error(1, "ex1"))),
            on_next(105, scheduler.create_cold_observable(on_error(1, "ex2"))),
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 102),
                    on_next(90, 103),
                    on_next(110, 104),
                    on_next(190, 105),
                    on_next(440, 106),
                    on_completed(460),
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(
                    on_next(180, 202), on_next(190, 203), on_completed(205)
                ),
            ),
            on_next(
                550,
                scheduler.create_cold_observable(
                    on_next(10, 301),
                    on_next(50, 302),
                    on_next(70, 303),
                    on_next(260, 304),
                    on_next(310, 305),
                    on_completed(410),
                ),
            ),
            on_next(750, scheduler.create_cold_observable(on_completed(40))),
            on_next(
                850,
                scheduler.create_cold_observable(
                    on_next(80, 401), on_next(90, 402), on_completed(100)
                ),
            ),
            on_completed(900),
        )

        def create():
            return xs.pipe(ops.flat_map(lambda x: x))

        results = scheduler.start(create, disposed=700)

        assert results.messages == [
            on_next(310, 102),
            on_next(390, 103),
            on_next(410, 104),
            on_next(490, 105),
            on_next(560, 301),
            on_next(580, 202),
            on_next(590, 203),
            on_next(600, 302),
            on_next(620, 303),
        ]
        assert xs.subscriptions == [subscribe(200, 700)]
        assert xs.messages[2].value.value.subscriptions == [subscribe(300, 700)]
        assert xs.messages[3].value.value.subscriptions == [subscribe(400, 605)]
        assert xs.messages[4].value.value.subscriptions == [subscribe(550, 700)]
        assert xs.messages[5].value.value.subscriptions == []
        assert xs.messages[6].value.value.subscriptions == []

    def test_flat_map_on_error(self):
        invoked = [0]
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(5, scheduler.create_cold_observable(on_error(1, "ex1"))),
            on_next(105, scheduler.create_cold_observable(on_error(1, "ex2"))),
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 102),
                    on_next(90, 103),
                    on_next(110, 104),
                    on_next(190, 105),
                    on_next(440, 106),
                    on_completed(460),
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(
                    on_next(180, 202), on_next(190, 203), on_completed(205)
                ),
            ),
            on_next(
                550,
                scheduler.create_cold_observable(
                    on_next(10, 301),
                    on_next(50, 302),
                    on_next(70, 303),
                    on_next(260, 304),
                    on_next(310, 305),
                    on_completed(410),
                ),
            ),
            on_next(750, scheduler.create_cold_observable(on_completed(40))),
            on_next(
                850,
                scheduler.create_cold_observable(
                    on_next(80, 401), on_next(90, 402), on_completed(100)
                ),
            ),
            on_completed(900),
        )

        def factory():
            def projection(x):
                invoked[0] += 1
                if invoked[0] == 3:
                    raise Exception(ex)
                return x

            return xs.pipe(ops.flat_map(projection))

        results = scheduler.start(factory)

        assert results.messages == [
            on_next(310, 102),
            on_next(390, 103),
            on_next(410, 104),
            on_next(490, 105),
            on_error(550, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 550)]
        assert xs.messages[2].value.value.subscriptions == [subscribe(300, 550)]
        assert xs.messages[3].value.value.subscriptions == [subscribe(400, 550)]
        assert xs.messages[4].value.value.subscriptions == []
        assert xs.messages[5].value.value.subscriptions == []
        assert xs.messages[6].value.value.subscriptions == []

    def test_flat_map_use_function(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(210, 4),
            on_next(220, 3),
            on_next(250, 5),
            on_next(270, 1),
            on_completed(290),
        )

        def factory():
            def projection(x):
                return reactivex.interval(10).pipe(
                    ops.map_indexed(lambda a, b: x), ops.take(x)
                )

            return xs.pipe(ops.flat_map(projection))

        results = scheduler.start(factory)

        assert results.messages == [
            on_next(220, 4),
            on_next(230, 3),
            on_next(230, 4),
            on_next(240, 3),
            on_next(240, 4),
            on_next(250, 3),
            on_next(250, 4),
            on_next(260, 5),
            on_next(270, 5),
            on_next(280, 1),
            on_next(280, 5),
            on_next(290, 5),
            on_next(300, 5),
            on_completed(300),
        ]
        assert xs.subscriptions == [subscribe(200, 290)]

    def test_flat_map_iterable_complete(self):
        scheduler = TestScheduler()

        xs = scheduler.create_hot_observable(
            on_next(210, 2),
            on_next(340, 4),
            on_next(420, 3),
            on_next(510, 2),
            on_completed(600),
        )
        inners = []

        def create():
            def mapper(x):
                ys = [x] * x
                inners.append(ys)
                return ys

            return xs.pipe(ops.flat_map(mapper))

        res = scheduler.start(create)

        assert res.messages == [
            on_next(210, 2),
            on_next(210, 2),
            on_next(340, 4),
            on_next(340, 4),
            on_next(340, 4),
            on_next(340, 4),
            on_next(420, 3),
            on_next(420, 3),
            on_next(420, 3),
            on_next(510, 2),
            on_next(510, 2),
            on_completed(600),
        ]

        assert xs.subscriptions == [subscribe(200, 600)]
        assert 4 == len(inners)


if __name__ == "__main__":
    unittest.main()



from datetime import datetime, timedelta
from time import sleep

import pytest

from reactivex.scheduler.eventloop import TwistedScheduler

twisted = pytest.importorskip("twisted")
from twisted.internet import defer, reactor  # isort: skip
from twisted.trial import unittest  # isort: skip


class TestTwistedScheduler(unittest.TestCase):
    def test_twisted_schedule_now(self):
        scheduler = TwistedScheduler(reactor)
        diff = scheduler.now - datetime.utcfromtimestamp(float(reactor.seconds()))
        assert abs(diff) < timedelta(milliseconds=1)

    def test_twisted_schedule_now_units(self):
        scheduler = TwistedScheduler(reactor)
        diff = scheduler.now
        sleep(0.1)
        diff = scheduler.now - diff
        assert timedelta(milliseconds=80) < diff < timedelta(milliseconds=180)

    @defer.inlineCallbacks
    def test_twisted_schedule_action(self):
        scheduler = TwistedScheduler(reactor)
        promise = defer.Deferred()
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        def done():
            promise.callback("Done")

        scheduler.schedule(action)
        reactor.callLater(0.1, done)

        yield promise
        assert ran is True

    @defer.inlineCallbacks
    def test_twisted_schedule_action_due(self):
        scheduler = TwistedScheduler(reactor)
        promise = defer.Deferred()
        starttime = reactor.seconds()
        endtime = None

        def action(scheduler, state):
            nonlocal endtime
            endtime = reactor.seconds()

        def done():
            promise.callback("Done")

        scheduler.schedule_relative(0.2, action)
        reactor.callLater(0.3, done)

        yield promise
        diff = endtime - starttime
        assert diff > 0.18

    @defer.inlineCallbacks
    def test_twisted_schedule_action_cancel(self):
        scheduler = TwistedScheduler(reactor)
        promise = defer.Deferred()
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        def done():
            promise.callback("Done")

        d = scheduler.schedule_relative(0.01, action)
        d.dispose()

        reactor.callLater(0.1, done)
        yield promise
        assert ran is False



import unittest

from reactivex.internal import PriorityQueue


class TestItem:
    __test__ = False

    def __init__(self, value, label=None):
        self.value = value
        self.label = label

    def __str__(self):
        if self.label:
            return "%s (%s)" % (self.value, self.label)
        else:
            return "%s" % self.value

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.value == other.value  # and self.label == other.label

    def __lt__(self, other):
        return self.value < other.value

    def __gt__(self, other):
        return self.value > other.value


class TestPriorityQueue(unittest.TestCase):
    def test_priorityqueue_count(self):
        assert PriorityQueue.MIN_COUNT < 0

    def test_priorityqueue_empty(self):
        """Must be empty on construction"""

        p = PriorityQueue()
        assert len(p) == 0
        assert p.items == []

        # Still empty after enqueue/dequeue
        p.enqueue(42)
        p.dequeue()
        assert len(p) == 0

    def test_priorityqueue_length(self):
        """Test that length is n after n invocations"""

        p = PriorityQueue()

        assert len(p) == 0
        for n in range(42):
            p.enqueue(n)
        assert len(p) == 42
        p.dequeue()
        assert len(p) == 41
        p.remove(10)
        assert len(p) == 40
        for n in range(len(p)):
            p.dequeue()
        assert len(p) == 0

    def test_priorityqueue_enqueue_dequeue(self):
        """Enqueue followed by dequeue should give the same result"""

        p = PriorityQueue()
        self.assertRaises(IndexError, p.dequeue)

        p.enqueue(42)
        p.enqueue(41)
        p.enqueue(43)

        assert [p.dequeue(), p.dequeue(), p.dequeue()] == [41, 42, 43]

    def test_priorityqueue_sort_stability(self):
        """Items with same value should be returned in the order they were
        added"""

        p = PriorityQueue()

        p.enqueue(TestItem(43, "high"))
        p.enqueue(TestItem(42, "first"))
        p.enqueue(TestItem(42, "second"))
        p.enqueue(TestItem(42, "last"))
        p.enqueue(TestItem(41, "low"))

        assert len(p) == 5

        assert p.dequeue() == TestItem(41, "low")
        assert p.dequeue() == TestItem(42, "first")
        assert p.dequeue() == TestItem(42, "second")
        assert p.dequeue() == TestItem(42, "last")
        assert p.dequeue() == TestItem(43, "high")

    def test_priorityqueue_remove(self):
        """Remove item from queue"""

        p = PriorityQueue()
        assert p.remove(42) == False

        p.enqueue(42)
        p.enqueue(41)
        p.enqueue(43)
        assert p.remove(42) == True
        assert [p.dequeue(), p.dequeue()] == [41, 43]

        p.enqueue(42)
        p.enqueue(41)
        p.enqueue(43)
        assert p.remove(41) == True
        assert [p.dequeue(), p.dequeue()] == [42, 43]

        p.enqueue(42)
        p.enqueue(41)
        p.enqueue(43)
        assert p.remove(43) == True
        assert [p.dequeue(), p.dequeue()] == [41, 42]

    def test_priorityqueue_peek(self):
        """Peek at first element in queue"""

        p = PriorityQueue()

        self.assertRaises(IndexError, p.peek)
        p.enqueue(42)
        assert p.peek() == 42
        p.enqueue(41)
        assert p.peek() == 41
        p.enqueue(43)
        assert p.peek() == 41



import unittest

from reactivex import operators as _
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestCount(unittest.TestCase):
    def test_count_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(250))
        res = scheduler.start(create=lambda: xs.pipe(_.count())).messages
        assert res == [on_next(250, 0), on_completed(250)]

    def test_count_empty_ii(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_completed(250)
        )

        def create():
            return xs.pipe(_.count())

        res = scheduler.start(create=create).messages
        assert res == [on_next(250, 1), on_completed(250)]

    def test_count_some(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_completed(250),
        )
        res = scheduler.start(create=lambda: xs.pipe(_.count())).messages
        assert res == [on_next(250, 3), on_completed(250)]

    def test_count_on_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_error(210, ex))
        res = scheduler.start(create=lambda: xs.pipe(_.count())).messages
        assert res == [on_error(210, ex)]

    def test_count_never(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1))
        res = scheduler.start(create=lambda: xs.pipe(_.count())).messages
        assert res == []

    def test_count_predicate_empty_true(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(250))

        def create():
            return xs.pipe(_.count(lambda _: True))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 0), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_count_predicate_empty_false(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(250))

        def create():
            return xs.pipe(_.count(lambda _: False))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 0), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_count_predicate_return_true(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_completed(250)
        )

        def create():
            return xs.pipe(_.count(lambda _: True))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 1), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_count_predicate_return_false(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_completed(250)
        )

        def create():
            return xs.pipe(_.count(lambda _: False))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 0), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_count_predicate_some_all(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_completed(250),
        )

        def create():
            return xs.pipe(_.count(lambda x: x < 10))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 3), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_count_predicate_some_none(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_completed(250),
        )

        def create():
            return xs.pipe(_.count(lambda x: x > 10))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 0), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_count_predicate_some_even(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_completed(250),
        )

        def create():
            return xs.pipe(_.count(lambda x: x % 2 == 0))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 2), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_count_predicate_throw_true(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_error(210, ex))

        def create():
            return xs.pipe(_.count(lambda _: True))

        res = scheduler.start(create=create)

        assert res.messages == [on_error(210, ex)]
        assert xs.subscriptions == [subscribe(200, 210)]

    def test_count_predicate_throw_false(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_error(210, ex))

        def create():
            return xs.pipe(_.count(lambda _: False))

        res = scheduler.start(create=create)

        assert res.messages == [on_error(210, ex)]
        assert xs.subscriptions == [subscribe(200, 210)]

    def test_count_predicate_never(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1))

        def create():
            return xs.pipe(_.count(lambda _: True))

        res = scheduler.start(create=create)

        assert res.messages == []
        assert xs.subscriptions == [subscribe(200, 1000)]

    def test_count_predicate_predicate_throws(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_next(230, 3), on_completed(240)
        )

        def create():
            def predicate(x):
                if x == 3:
                    raise Exception(ex)
                else:
                    return True

            return xs.pipe(_.count(predicate))

        res = scheduler.start(create=create)

        assert res.messages == [on_error(230, ex)]
        assert xs.subscriptions == [subscribe(200, 230)]



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(ex):
    raise RxException(ex)


class TestThrottleFirst(unittest.TestCase):
    def test_throttle_first_completed(self):
        scheduler = TestScheduler()

        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(310, 4),
            on_next(350, 5),
            on_next(410, 6),
            on_next(450, 7),
            on_completed(500),
        )

        def create():
            return xs.pipe(ops.throttle_first(200))

        results = scheduler.start(create=create)

        assert results.messages == [on_next(210, 2), on_next(410, 6), on_completed(500)]

        assert xs.subscriptions == [subscribe(200, 500)]

    def test_throttle_first_never(self):
        scheduler = TestScheduler()

        xs = scheduler.create_hot_observable(on_next(150, 1))

        def create():
            return xs.pipe(ops.throttle_first(200))

        results = scheduler.start(create=create)

        assert results.messages == []
        assert xs.subscriptions == [subscribe(200, 1000)]

    def test_throttle_first_empty(self):
        scheduler = TestScheduler()

        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(500))

        def create():
            return xs.pipe(ops.throttle_first(200))

        results = scheduler.start(create=create)

        assert results.messages == [on_completed(500)]

        assert xs.subscriptions == [subscribe(200, 500)]

    def test_throttle_first_error(self):
        error = RxException()

        scheduler = TestScheduler()

        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(310, 4),
            on_next(350, 5),
            on_error(410, error),
            on_next(450, 7),
            on_completed(500),
        )

        def create():
            return xs.pipe(ops.throttle_first(200))

        results = scheduler.start(create=create)

        assert results.messages == [on_next(210, 2), on_error(410, error)]

        assert xs.subscriptions == [subscribe(200, 410)]

    def test_throttle_first_no_end(self):
        scheduler = TestScheduler()

        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(250, 3),
            on_next(310, 4),
            on_next(350, 5),
            on_next(410, 6),
            on_next(450, 7),
        )

        def create():
            return xs.pipe(ops.throttle_first(200))

        results = scheduler.start(create=create)

        assert results.messages == [on_next(210, 2), on_next(410, 6)]

        assert xs.subscriptions == [subscribe(200, 1000)]



import unittest
from datetime import timedelta
from time import sleep

import pytest

from reactivex.internal.basic import default_now
from reactivex.scheduler.mainloop import WxScheduler

wx = pytest.importorskip("wx")


def make_app():
    app = wx.App()
    wx.Frame(None)  # We need this for some reason, or the loop won't run
    return app


class AppExit(wx.Timer):
    def __init__(self, app) -> None:
        super().__init__()
        self.app = app

    def Notify(self):
        self.app.ExitMainLoop()


class TestWxScheduler(unittest.TestCase):
    def test_wx_schedule_now(self):
        scheduler = WxScheduler(wx)
        diff = scheduler.now - default_now()
        assert abs(diff) < timedelta(milliseconds=1)

    def test_wx_schedule_now_units(self):
        scheduler = WxScheduler(wx)
        diff = scheduler.now
        sleep(0.1)
        diff = scheduler.now - diff
        assert timedelta(milliseconds=80) < diff < timedelta(milliseconds=180)

    def test_wx_schedule_action(self):
        app = make_app()
        exit = AppExit(app)
        scheduler = WxScheduler(wx)
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        scheduler.schedule(action)
        exit.Start(100, wx.TIMER_ONE_SHOT)
        app.MainLoop()
        scheduler.cancel_all()

        assert ran is True

    def test_wx_schedule_action_relative(self):
        app = make_app()
        exit = AppExit(app)
        scheduler = WxScheduler(wx)
        starttime = default_now()
        endtime = None

        def action(scheduler, state):
            nonlocal endtime
            endtime = default_now()

        scheduler.schedule_relative(0.1, action)
        exit.Start(200, wx.TIMER_ONE_SHOT)
        app.MainLoop()
        scheduler.cancel_all()

        assert endtime is not None
        diff = endtime - starttime
        assert timedelta(milliseconds=80) < diff < timedelta(milliseconds=180)

    def test_wx_schedule_action_absolute(self):
        app = make_app()
        exit = AppExit(app)
        scheduler = WxScheduler(wx)
        starttime = default_now()
        endtime = None

        def action(scheduler, state):
            nonlocal endtime
            endtime = default_now()

        due = scheduler.now + timedelta(milliseconds=100)
        scheduler.schedule_absolute(due, action)
        exit.Start(200, wx.TIMER_ONE_SHOT)
        app.MainLoop()
        scheduler.cancel_all()

        assert endtime is not None
        diff = endtime - starttime
        assert timedelta(milliseconds=80) < diff < timedelta(milliseconds=180)

    def test_wx_schedule_action_cancel(self):
        app = make_app()
        exit = AppExit(app)
        scheduler = WxScheduler(wx)
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        d = scheduler.schedule_relative(0.1, action)
        d.dispose()
        exit.Start(200, wx.TIMER_ONE_SHOT)
        app.MainLoop()
        scheduler.cancel_all()

        assert ran is False

    def test_wx_schedule_action_periodic(self):
        app = make_app()
        exit = AppExit(app)
        scheduler = WxScheduler(wx)
        period = 0.05
        counter = 3

        def action(state):
            nonlocal counter
            if state:
                counter -= 1
                return state - 1

        scheduler.schedule_periodic(period, action, counter)
        exit.Start(500, wx.TIMER_ONE_SHOT)
        app.MainLoop()
        scheduler.cancel_all()

        assert counter == 0



import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(error: str):
    raise RxException(error)


class TestConcat(unittest.TestCase):
    def test_concat_empty_empty(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(230)]
        msgs2 = [on_next(150, 1), on_completed(250)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [on_completed(250)]

    def test_concat_empty_never(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(230)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = reactivex.never()

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == []

    def test_concat_never_empty(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(230)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = reactivex.never()

        def create():
            return e2.pipe(ops.concat(e1))

        results = scheduler.start(create)
        assert results.messages == []

    def test_concat_never_never(self):
        scheduler = TestScheduler()
        e1 = reactivex.never()
        e2 = reactivex.never()

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == []

    def test_concat_empty_on_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(230)]
        msgs2 = [on_next(150, 1), on_error(250, ex)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [on_error(250, ex)]

    def test_concat_throw_empty(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_error(230, ex)]
        msgs2 = [on_next(150, 1), on_completed(250)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [on_error(230, ex)]

    def test_concat_throw_on_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_error(230, ex)]
        msgs2 = [on_next(150, 1), on_error(250, "ex2")]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [on_error(230, ex)]

    def test_concat_return_empty(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_completed(230)]
        msgs2 = [on_next(150, 1), on_completed(250)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2), on_completed(250)]

    def test_concat_empty_return(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(230)]
        msgs2 = [on_next(150, 1), on_next(240, 2), on_completed(250)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [on_next(240, 2), on_completed(250)]

    def test_concat_return_never(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_completed(230)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = reactivex.never()

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2)]

    def test_concat_never_return(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_completed(230)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = reactivex.never()

        def create():
            return e2.pipe(ops.concat(e1))

        results = scheduler.start(create)
        assert results.messages == []

    def test_concat_return_return(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(220, 2), on_completed(230)]
        msgs2 = [on_next(150, 1), on_next(240, 3), on_completed(250)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [on_next(220, 2), on_next(240, 3), on_completed(250)]

    def test_concat_throw_return(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_error(230, ex)]
        msgs2 = [on_next(150, 1), on_next(240, 2), on_completed(250)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [on_error(230, ex)]

    def test_concat_return_on_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(220, 2), on_completed(230)]
        msgs2 = [on_next(150, 1), on_error(250, ex)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [on_next(220, 2), on_error(250, ex)]

    def test_concat_some_data_some_data(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_next(220, 3), on_completed(225)]
        msgs2 = [on_next(150, 1), on_next(230, 4), on_next(240, 5), on_completed(250)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        def create():
            return e1.pipe(ops.concat(e2))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        ]

    def test_concat_forward_scheduler(self):
        scheduler = TestScheduler()
        subscribe_schedulers = {"e1": "unknown", "e2": "unknown"}

        def subscribe_e1(observer, scheduler="not_set"):
            subscribe_schedulers["e1"] = scheduler
            observer.on_completed()

        def subscribe_e2(observer, scheduler="not_set"):
            subscribe_schedulers["e2"] = scheduler
            observer.on_completed()

        e1 = reactivex.create(subscribe_e1)
        e2 = reactivex.create(subscribe_e2)

        stream = e1.pipe(ops.concat(e2))
        stream.subscribe(scheduler=scheduler)
        scheduler.advance_to(1000)
        assert subscribe_schedulers["e1"] is scheduler
        assert subscribe_schedulers["e2"] is scheduler

    def test_concat_forward_none_scheduler(self):
        subscribe_schedulers = {"e1": "unknown", "e2": "unknown"}

        def subscribe_e1(observer, scheduler="not_set"):
            subscribe_schedulers["e1"] = scheduler
            observer.on_completed()

        def subscribe_e2(observer, scheduler="not_set"):
            subscribe_schedulers["e2"] = scheduler
            observer.on_completed()

        e1 = reactivex.create(subscribe_e1)
        e2 = reactivex.create(subscribe_e2)

        stream = e1.pipe(ops.concat(e2))
        stream.subscribe()
        assert subscribe_schedulers["e1"] is None
        assert subscribe_schedulers["e2"] is None



"""
This example shows how you can retry subscriptions that might
sometime fail with an error.

Note: you cannot use ref_count() in this case since that would
make publish() re-subscribe the cold-observable and it would start
looping forever.
"""

import time

import reactivex
from reactivex import operators as ops


def failing(x):
    x = int(x)
    if not x % 2:
        raise Exception("Error")
    return x


def main():
    xs = reactivex.from_marbles("1-2-3-4-5-6-7-9-|").pipe(ops.publish())
    xs.pipe(ops.map(failing), ops.retry()).subscribe(print)

    xs.connect()  # Must connect. Cannot use ref_count() with publish()

    time.sleep(5)


if __name__ == "__main__":
    main()



import datetime

import reactivex
import reactivex.operators as ops

"""
Delay the emission of elements to the specified datetime.
"""

now = datetime.datetime.utcnow()
dt = datetime.timedelta(seconds=3.0)
duetime = now + dt

print(
    "{} ->  now\n"
    "{} ->  start of emission in {}s".format(now, duetime, dt.total_seconds())
)

hot = reactivex.hot("10--11--12--13--(14,|)", timespan=0.2, duetime=duetime)

source = hot.pipe(ops.do_action(print))
source.run()



import wx

import reactivex
from reactivex import operators as ops
from reactivex.scheduler.mainloop import WxScheduler
from reactivex.subject import Subject


class Frame(wx.Frame):
    def __init__(self):
        super(Frame, self).__init__(None)
        self.SetTitle("Rx for Python rocks")
        self.SetSize((600, 600))

        # This Subject is used to transmit mouse moves to labels
        self.mousemove = Subject()

        self.Bind(wx.EVT_MOTION, self.OnMotion)

    def OnMotion(self, event):
        self.mousemove.on_next((event.GetX(), event.GetY()))


def main():
    app = wx.App()
    scheduler = WxScheduler(wx)

    app.TopWindow = frame = Frame()
    frame.Show()

    text = "TIME FLIES LIKE AN ARROW"

    def on_next(info):
        label, (x, y), i = info
        label.Move(x + i * 12 + 15, y)
        label.Show()

    def handle_label(label, i):
        delayer = ops.delay(i * 0.100)
        mapper = ops.map(lambda xy: (label, xy, i))

        return frame.mousemove.pipe(
            delayer,
            mapper,
        )

    def make_label(char):
        label = wx.StaticText(frame, label=char)
        label.Hide()
        return label

    mapper = ops.map(make_label)
    labeler = ops.flat_map_indexed(handle_label)

    reactivex.from_(text).pipe(
        mapper,
        labeler,
    ).subscribe(on_next, on_error=print, scheduler=scheduler)

    frame.Bind(wx.EVT_CLOSE, lambda e: (scheduler.cancel_all(), e.Skip()))
    app.MainLoop()


if __name__ == "__main__":
    main()



from reactivex import operators as ops
from reactivex.testing.marbles import marbles_testing

"""
Tests MergeMap from reactivexjs
https://github.com/ReactiveX/rxjs/blob/master/spec/operators/mergeMap-spec.ts

it should flat_map many regular interval inners
"""
with marbles_testing(timespan=1.0) as context:
    start, cold, hot, exp = context

    a = cold(" ----a---a----a----(a,|)                     ")
    b = cold("     ----1----b----(b,|)                     ")
    c = cold("                 -------c---c---c----c---(c,|)")
    d = cold("                         -------(d,|)        ")
    e1 = hot("-a---b-----------c-------d------------|      ")
    ex = exp("-----a---(a,1)(a,b)(a,b)c---c---(c,d)c---(c,|)")
    expected = ex

    observableLookup = {"a": a, "b": b, "c": c, "d": d}

    obs = e1.pipe(ops.flat_map(lambda value: observableLookup[value]))

    results = start(obs)
    assert results == expected

print("flat_map: results vs expected")
for r, e in zip(results, expected):
    print(r, e)



"""
RxPY example running a Tornado server doing search queries against
Wikipedia to populate the autocomplete dropdown in the web UI. Start
using `python autocomplete.py` and navigate your web browser to
http://localhost:8080

Uses the RxPY AsyncIOScheduler (Python 3.4 is required)
"""

import asyncio
import os
from asyncio import Future
from typing import Dict, Union

from tornado.escape import json_decode
from tornado.httpclient import AsyncHTTPClient, HTTPResponse
from tornado.httputil import url_concat
from tornado.platform.asyncio import AsyncIOMainLoop
from tornado.web import Application, RequestHandler, StaticFileHandler, url
from tornado.websocket import WebSocketHandler

from reactivex import operators as ops
from reactivex.scheduler.eventloop import AsyncIOScheduler
from reactivex.subject import Subject


def search_wikipedia(term: str) -> Future[HTTPResponse]:
    """Search Wikipedia for a given term"""
    url = "http://en.wikipedia.org/w/api.php"

    params = {"action": "opensearch", "search": term, "format": "json"}
    # Must set a user agent for non-browser requests to Wikipedia
    user_agent = (
        "RxPY/1.0 (https://github.com/dbrattli/RxPY; dag@brattli.net) Tornado/4.0.1"
    )

    url = url_concat(url, params)

    http_client = AsyncHTTPClient()
    return http_client.fetch(url, method="GET", user_agent=user_agent)


class WSHandler(WebSocketHandler):
    def open(self):
        scheduler = AsyncIOScheduler(asyncio.get_event_loop())

        print("WebSocket opened")

        # A Subject is both an observable and observer, so we can both subscribe
        # to it and also feed (send) it with new values
        self.subject: Subject[Dict[str, str]] = Subject()

        # Get all distinct key up events from the input and only fire if long enough and distinct
        searcher = self.subject.pipe(
            ops.map(lambda x: x["term"]),
            ops.filter(
                lambda text: len(text) > 2
            ),  # Only if the text is longer than 2 characters
            ops.debounce(0.750),  # Pause for 750ms
            ops.distinct_until_changed(),  # Only if the value has changed
            ops.flat_map_latest(search_wikipedia),
        )

        def send_response(x: HTTPResponse) -> None:
            self.write_message(x.body)

        def on_error(ex: Exception):
            print(ex)

        searcher.subscribe(
            on_next=send_response, on_error=on_error, scheduler=scheduler
        )

    def on_message(self, message: Union[bytes, str]):
        obj = json_decode(message)
        self.subject.on_next(obj)

    def on_close(self):
        print("WebSocket closed")


class MainHandler(RequestHandler):
    def get(self):
        self.render("index.html")


def main():
    AsyncIOMainLoop().make_current()

    port = os.environ.get("PORT", 8080)
    app = Application(
        [
            url(r"/", MainHandler),
            (r"/ws", WSHandler),
            (r"/static/(.*)", StaticFileHandler, {"path": "."}),
        ]
    )

    print("Starting server at port: %s" % port)
    app.listen(port)
    asyncio.get_event_loop().run_forever()


if __name__ == "__main__":
    main()



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler
from reactivex.testing.marbles import marbles_testing

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestBuffer(unittest.TestCase):
    def test_buffer_simple(self):
        scheduler = TestScheduler()

        xs = scheduler.create_hot_observable(
            on_next(90, 1),
            on_next(180, 2),
            on_next(250, 3),
            on_next(260, 4),
            on_next(310, 5),
            on_next(340, 6),
            on_next(410, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_next(550, 10),
            on_completed(590),
        )

        ys = scheduler.create_hot_observable(
            on_next(255, True),
            on_next(330, True),
            on_next(350, True),
            on_next(400, True),
            on_next(500, True),
            on_completed(900),
        )

        def create():
            return xs.pipe(ops.buffer(ys))

        res = scheduler.start(create=create)

        assert [
            on_next(255, lambda b: b == [3]),
            on_next(330, lambda b: b == [4, 5]),
            on_next(350, lambda b: b == [6]),
            on_next(400, lambda b: b == []),
            on_next(500, lambda b: b == [7, 8, 9]),
            on_next(590, lambda b: b == [10]),
            on_completed(590),
        ] == res.messages

        assert xs.subscriptions == [subscribe(200, 590)]

        assert ys.subscriptions == [subscribe(200, 590)]

    def test_buffer_closeboundaries(self):
        scheduler = TestScheduler()

        xs = scheduler.create_hot_observable(
            on_next(90, 1),
            on_next(180, 2),
            on_next(250, 3),
            on_next(260, 4),
            on_next(310, 5),
            on_next(340, 6),
            on_next(410, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_next(550, 10),
            on_completed(590),
        )

        ys = scheduler.create_hot_observable(
            on_next(255, True),
            on_next(330, True),
            on_next(350, True),
            on_completed(400),
        )

        def create():
            return xs.pipe(ops.buffer(ys))

        res = scheduler.start(create=create)

        assert [
            on_next(255, lambda b: b == [3]),
            on_next(330, lambda b: b == [4, 5]),
            on_next(350, lambda b: b == [6]),
            on_next(400, lambda b: b == []),
            on_completed(400),
        ] == res.messages

        assert xs.subscriptions == [subscribe(200, 400)]

        assert ys.subscriptions == [subscribe(200, 400)]

    def test_buffer_throwsource(self):
        ex = "ex"

        scheduler = TestScheduler()

        xs = scheduler.create_hot_observable(
            on_next(90, 1),
            on_next(180, 2),
            on_next(250, 3),
            on_next(260, 4),
            on_next(310, 5),
            on_next(340, 6),
            on_next(380, 7),
            on_error(400, ex),
        )

        ys = scheduler.create_hot_observable(
            on_next(255, True),
            on_next(330, True),
            on_next(350, True),
            on_completed(500),
        )

        def create():
            return xs.pipe(ops.buffer(ys))

        res = scheduler.start(create=create)

        assert [
            on_next(255, lambda b: b == [3]),
            on_next(330, lambda b: b == [4, 5]),
            on_next(350, lambda b: b == [6]),
            on_error(400, ex),
        ] == res.messages

        assert xs.subscriptions == [subscribe(200, 400)]

        assert ys.subscriptions == [subscribe(200, 400)]

    def test_buffer_throwboundaries(self):
        ex = "ex"

        scheduler = TestScheduler()

        xs = scheduler.create_hot_observable(
            on_next(90, 1),
            on_next(180, 2),
            on_next(250, 3),
            on_next(260, 4),
            on_next(310, 5),
            on_next(340, 6),
            on_next(410, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_next(550, 10),
            on_completed(590),
        )

        ys = scheduler.create_hot_observable(
            on_next(255, True),
            on_next(330, True),
            on_next(350, True),
            on_error(400, ex),
        )

        def create():
            return xs.pipe(ops.buffer(ys))

        res = scheduler.start(create=create)

        assert [
            on_next(255, lambda b: b == [3]),
            on_next(330, lambda b: b == [4, 5]),
            on_next(350, lambda b: b == [6]),
            on_error(400, ex),
        ] == res.messages

        assert xs.subscriptions == [subscribe(200, 400)]

        assert ys.subscriptions == [subscribe(200, 400)]

    def test_when_closing_with_empty_observable(self):
        with marbles_testing(timespan=1.0) as (start, cold, hot, exp):

            def closing_mapper():
                return cold("-----|")

            lookup = {"a": [1, 2], "b": [3], "c": [4, 5], "d": [6], "e": [7]}
            #               -----|
            #                    -----|
            #                         -----|
            #                              -----|
            #                                   -----|
            source = hot("  -1--2---3--4--5--6---7--|")
            expected = exp("-----a----b----c----d---(e,|)", lookup=lookup)
            #               012345678901234567890123456789
            #               0         1         2
            obs = source.pipe(ops.buffer_when(closing_mapper))
            results = start(obs)
            assert results == expected

    def test_when_closing_only_on_first_item(self):
        with marbles_testing(timespan=1.0) as (start, cold, hot, exp):

            def closing_mapper():
                return cold("-----1--2--3--4--|")

            lookup = {"a": [1, 2], "b": [3], "c": [4, 5], "d": [6], "e": [7]}
            #               -----1--2--3--4--|
            #                    -----1--2--3--4--|
            #                         -----1--2--3--4--|
            #                              -----1--2--3--4--|
            source = hot("  -1--2---3--4--5--6---7--|")
            expected = exp("-----a----b----c----d---(e,|)", lookup=lookup)
            #               012345678901234567890123456789
            #               0         1         2
            obs = source.pipe(ops.buffer_when(closing_mapper))
            results = start(obs)
            assert results == expected

    def test_when_source_on_error(self):
        class TestException(Exception):
            pass

        ex = TestException("test exception")

        with marbles_testing(timespan=1.0) as (start, cold, hot, exp):

            def closing_mapper():
                return cold("-----1|")

            lookup = {"a": [1, 2], "b": [3], "c": [4, 5], "d": [6], "e": [7]}
            #               -----1|
            #                    -----1|
            #                         -----1|
            #                              -----1|
            #                                   -----1|
            source = hot("  -1--2---3--4--5--#", error=ex)
            expected = exp("-----a----b----c-#", lookup=lookup, error=ex)
            #               012345678901234567890123456789
            #               0         1         2

            obs = source.pipe(ops.buffer_when(closing_mapper))
            results = start(obs)
            assert results == expected

    def test_when_closing_on_error(self):
        class TestException(Exception):
            pass

        ex = TestException("test exception")

        with marbles_testing(timespan=1.0) as (start, cold, hot, exp):

            def closing_mapper():
                return cold("-----#", error=ex)

            #               -----#
            source = hot("  -1--2---3--4--5--6---7--|")
            expected = exp("-----#", error=ex)
            #               012345678901234567890123456789
            #               0         1         2
            obs = source.pipe(ops.buffer_when(closing_mapper))
            results = start(obs)
            assert results == expected

    def test_toggle_closing_with_empty_observable(self):
        with marbles_testing(timespan=1.0) as (start, cold, hot, exp):

            lookup = {"a": [2, 3], "b": [5], "c": [7], "d": [8, 9]}

            openings = hot("---a-------b------c---d-------|")
            a = cold("         ------|")
            b = cold("                 ----|")
            c = cold("                        ---|")
            d = cold("                            -----|")
            source = hot("  -1--2--3--4--5--6---7--8--9---|")
            expected = exp("---------a-----b-----c-----d--|", lookup=lookup)
            #               012345678901234567890123456789
            #               0         1         2
            closings = {"a": a, "b": b, "c": c, "d": d}

            def closing_mapper(key):
                return closings[key]

            obs = source.pipe(ops.buffer_toggle(openings, closing_mapper))
            results = start(obs)
            assert results == expected

    def test_toggle_closing_with_only_first_item(self):
        with marbles_testing(timespan=1.0) as (start, cold, hot, exp):

            lookup = {"a": [2, 3], "b": [5], "c": [7], "d": [8, 9]}

            openings = hot("---a-------b------c---d-------|")
            a = cold("         ------1-2-|")
            b = cold("                 ----1-2-|")
            c = cold("                        ---1-2-|")
            d = cold("                            -----1-2|")
            source = hot("  -1--2--3--4--5--6---7--8--9---|")
            expected = exp("---------a-----b-----c-----d--|", lookup=lookup)
            #               012345678901234567890123456789
            #               0         1         2
            closings = {"a": a, "b": b, "c": c, "d": d}

            def closing_mapper(key):
                return closings[key]

            obs = source.pipe(ops.buffer_toggle(openings, closing_mapper))
            results = start(obs)
            assert results == expected

    def test_toggle_source_on_error(self):
        class TestException(Exception):
            pass

        ex = TestException("test exception")

        with marbles_testing(timespan=1.0) as (start, cold, hot, exp):

            lookup = {"a": [2, 3], "b": [5], "c": [7], "d": [8, 9]}

            openings = hot("---a-------b------c---d-------|")
            a = cold("         ------1-2-|")
            b = cold("                 ----1-2-|")
            c = cold("                        ---1-2-|")
            d = cold("                            -----1-2|")
            source = hot("  -1--2--3--4--5--6--#", error=ex)
            expected = exp("---------a-----b---#", lookup=lookup, error=ex)
            #               012345678901234567890123456789
            #               0         1         2
            closings = {"a": a, "b": b, "c": c, "d": d}

            def closing_mapper(key):
                return closings[key]

            obs = source.pipe(ops.buffer_toggle(openings, closing_mapper))
            results = start(obs)
            assert results == expected

    def test_toggle_openings_on_error(self):
        class TestException(Exception):
            pass

        ex = TestException("test exception")

        with marbles_testing(timespan=1.0) as (start, cold, hot, exp):

            lookup = {"a": [2, 3], "b": [5], "c": [7], "d": [8, 9]}

            openings = hot("---a-------b-----#", error=ex)
            a = cold("         ------1|")
            b = cold("                 ----1|")
            c = cold("                        ---1|")
            d = cold("                            -----1|")
            source = hot("  -1--2--3--4--5--6---7--8--9---|")
            expected = exp("---------a-----b-#", lookup=lookup, error=ex)
            #               012345678901234567890123456789
            #               0         1         2
            closings = {"a": a, "b": b, "c": c, "d": d}

            def closing_mapper(key):
                return closings[key]

            obs = source.pipe(ops.buffer_toggle(openings, closing_mapper))
            results = start(obs)
            assert results == expected

    def test_toggle_closing_mapper_on_error(self):
        class TestException(Exception):
            pass

        ex = TestException("test exception")

        with marbles_testing(timespan=1.0) as (start, cold, hot, exp):

            lookup = {"a": [2, 3], "b": [5], "c": [7], "d": [8, 9]}

            openings = hot("---a-------b------c---d-------|")
            a = cold("         ------1|")
            b = cold("                 ---#", error=ex)
            c = cold("                        ---1|")
            d = cold("                            -----1|")
            source = hot("  -1--2--3--4--5--6---7--8--9---|")
            expected = exp("---------a----#", lookup=lookup, error=ex)
            #               012345678901234567890123456789
            #               0         1         2
            closings = {"a": a, "b": b, "c": c, "d": d}

            def closing_mapper(key):
                return closings[key]

            obs = source.pipe(ops.buffer_toggle(openings, closing_mapper))
            results = start(obs)
            assert results == expected



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestSubscribeOn(unittest.TestCase):
    def test_subscribe_on_normal(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_completed(250),
        )

        def create():
            return xs.pipe(ops.subscribe_on(scheduler))

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_subscribe_on_error(self):
        scheduler = TestScheduler()
        ex = "ex"
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_error(210, ex),
        )

        def create():
            return xs.pipe(ops.subscribe_on(scheduler))

        results = scheduler.start(create)

        assert results.messages == [on_error(210, ex)]
        assert xs.subscriptions == [subscribe(200, 210)]

    def test_subscribe_on_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_completed(250),
        )

        def create():
            return xs.pipe(ops.subscribe_on(scheduler))

        results = scheduler.start(create)

        assert results.messages == [on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_subscribe_on_never(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1))

        def create():
            return xs.pipe(ops.subscribe_on(scheduler))

        results = scheduler.start(create)

        assert results.messages == []
        assert xs.subscriptions == [subscribe(200, 1000)]


if __name__ == "__main__":
    unittest.main()
